import xml.etree.ElementTree as ET
from contextlib import closing
from uuid import uuid4

from pyjabber.db.database import connection
from pyjabber.metadata import host
from pyjabber.stanzas.error import StanzaError as SE
from pyjabber.stanzas.IQ import IQ
from pyjabber.stream.JID import JID
from pyjabber.utils import Singleton


class Roster(metaclass=Singleton):
    """
    Roster plugin.

    Manages the roster list for each registered user in the server.

    A roster in XMPP is a server-stored contact list that manages contacts, presence status,
    and subscription requests.
    It enables real-time presence updates, contact organization, and synchronization across devices,
    ensuring seamless and private communication.

    :param db_connection_factory: The DB connection object
    """

    def __init__(self) -> None:
        self._handlers = {
            "get": self.handle_get,
            "set": self.handle_set,
            "result": self.handle_result
        }
        self._ns = {
            "ns": "jabber:iq:roster",
            "query": "{jabber:iq:roster}query",
            "item": "{jabber:iq:roster}item"
        }
        self._connection = connection

        self._roster_in_memory = {}
        self._update_roster()

    def create_roster_entry(self, jid: JID,  to: JID):
        iq = IQ(from_=str(jid), type_=IQ.TYPE.SET)
        query = ET.SubElement(iq, "{jabber:iq:roster}query")
        if to.domain == host.get():
            ET.SubElement(query, "{jabber:iq:roster}item", attrib={"jid": to.user, "subscription": "none"})
        else:
            ET.SubElement(query, "{jabber:iq:roster}item", attrib={"jid": to.bare(), "subscription": "none"})

        return self.feed(jid, iq)

    def check_pending_sub_to(self, jid: JID, to: str) -> ET.Element:
        with closing(connection()) as con:
            res = con.execute("SELECT * FROM pendingsub WHERE jid = ?", (jid,))
            res = res.fetchone()
        if res:
            return res

    def store_pending_sub(self, to_: str, item: ET.Element) -> None:
        with closing(connection()) as con:
            con.execute("INSERT INTO pendingsub values (?, ?)", (to_, ET.tostring(item).decode()))
            con.commit()

    def update_item(self, item: ET.Element, id_: int):
        with closing(connection()) as con:
            con.execute("UPDATE roster SET rosterItem = ? WHERE id = ?",
                        (ET.tostring(item).decode(), id_))
            con.commit()
        self._update_roster()

    def _update_roster(self):
        with closing(self._connection()) as con:
            res = con.execute("SELECT id, jid, rosterItem FROM roster", ())
            res = res.fetchall()
        self._roster_in_memory.clear()
        for id_, jid, item in res:
            if jid not in self._roster_in_memory:
                self._roster_in_memory[jid] = []
            self._roster_in_memory[jid].append({"id": id_, "item": item})

    def roster_by_jid(self, jid: JID):
        if jid.domain == host.get():
            return self._roster_in_memory.get(jid.user) or []
        return self._roster_in_memory.get(jid.bare()) or []

    def feed(self, jid: JID, element: ET.Element):
        if len(element) != 1:
            return SE.invalid_xml()

        return self._handlers[element.attrib.get("type")](jid, element)

    def handle_get(self, jid: JID, element: ET.Element):
        if jid.domain == host.get():
            jid = jid.user
        else:
            jid = jid.bare()

        roster = self._roster_in_memory.get(jid)

        iq = IQ(type_=IQ.TYPE.RESULT, id_=element.attrib.get("id"))
        query = ET.SubElement(iq, "query", attrib={"xmlns": "jabber:iq:roster"})

        for item in roster or []:
            query.append(ET.fromstring(item.get("item")))

        return ET.tostring(iq)

    def handle_set(self, jid: JID, element: ET.Element):
        query = element.find("{jabber:iq:roster}query")
        if jid.domain == host.get():
            jid = jid.user
        else:
            jid = jid.bare()

        if query is None:
            return SE.invalid_xml()

        new_item = query.findall("{jabber:iq:roster}item")

        if len(new_item) != 1:
            return SE.invalid_xml()

        new_item = new_item[0]
        # remove = None
        # if "subscription" in new_item.attrib.keys():
        #     remove = new_item.attrib["subscription"] == "remove"

        roster = self._roster_in_memory.get(jid)
        if roster:
            match_item = [i for i in roster if ET.fromstring(i.get("item")).get("jid") == new_item.attrib.get("jid")]
            if match_item:
                match_item = match_item[0]
                if new_item.attrib.get("remove") == "remove":
                    with closing(self._connection()) as con:
                        con.execute("DELETE FROM roster WHERE jid = ? AND rosterItem = ?",
                                    (jid, ET.tostring(match_item.get("item")).decode()))
                        con.commit()
                else:
                    with closing(self._connection()) as con:
                        con.execute("UPDATE roster SET rosterItem = ? WHERE jid = ? AND rosterItem = ?",
                                    (ET.tostring(new_item).decode(), jid, match_item.get("item")))
                        con.commit()
            else:
                if new_item.attrib.get("remove") != "remove":
                    with closing(self._connection()) as con:
                        con.execute("INSERT INTO roster(jid, rosterItem) VALUES (?, ?)",
                                    (jid, ET.tostring(new_item).decode()))
                        con.commit()

        else:
            with closing(self._connection()) as con:
                con.execute("INSERT INTO roster(jid, rosterItem) VALUES (?, ?)",
                            (jid, ET.tostring(new_item).decode()))
                con.commit()

        self._update_roster()
        res = IQ(id_=element.attrib.get("id"), type_=IQ.TYPE.RESULT)
        return ET.tostring(res)

    def handle_result(self, _, __):
        # It's safe to ignore this stanza
        return



import pytest
from unittest.mock import Mock, MagicMock, patch
from xml.etree import ElementTree as ET

from pyjabber.utils import ClarkNotation as CN
from pyjabber.stream.StreamHandler import Signal
from pyjabber.network.XMLParser import XMLParser


@pytest.fixture
def setup():
    with patch('pyjabber.stream.StreamHandler.host') as mock_host:
        buffer = Mock()
        starttls = Mock()
        mock_host.get.return_value = 'localhost'

        return XMLParser(buffer, starttls)

def test_initialization(setup):
    handler = setup

    assert handler._state == XMLParser.StreamState.CONNECTED
    assert handler._buffer is not None
    assert handler._stanzaHandler is None
    assert handler._streamHandler is not None
    assert handler._stack == []

def test_buffer_property(setup):
    handler = setup

    new_buffer = Mock()
    handler.buffer = new_buffer

    assert handler.buffer == new_buffer
    assert handler._streamHandler.buffer == new_buffer


def test_start_element_ns_stream(setup):
    handler = setup

    attrs = {("http://etherx.jabber.org/streams", "version"): "1.0"}
    handler.startElementNS(("http://etherx.jabber.org/streams", "stream"), "stream", attrs)

    assert handler._buffer.write.call_count == 2
    assert len(handler._stack) == 1
    assert handler._stack[0].tag == CN.clarkFromTuple(("http://etherx.jabber.org/streams", "stream"))


def test_start_element_ns_normal_element(setup):
    handler = setup

    attrs = {("namespace", "attr"): "value"}
    handler._stack.append(ET.Element("dummy"))

    handler.startElementNS(("namespace", "element"), "element", attrs)
    assert len(handler._stack) == 2
    assert handler._stack[-1].tag == CN.clarkFromTuple(("namespace", "element"))
    assert handler._stack[-1].attrib == {CN.clarkFromTuple(("namespace", "attr")): "value"}


def test_start_element_ns_invalid(setup):
    handler = setup

    with pytest.raises(Exception):
        handler.startElementNS(("invalid", "element"), "element", {})


def test_end_element_ns_stream(setup):
    handler = setup
    handler._stack.append(ET.Element("{http://etherx.jabber.org/streams}stream"))

    handler.endElementNS(("http://etherx.jabber.org/streams", "stream"), "stream")

    handler._buffer.write.assert_called_once_with(b'</stream:stream>')
    assert len(handler._stack) == 0


def test_end_element_ns_normal_element(setup):
    handler = setup

    parent = ET.Element(CN.clarkFromTuple(("namespace", "parent")))
    child = ET.Element(CN.clarkFromTuple(("namespace", "child")))

    handler._stack.append(parent)
    handler._stack.append(child)

    handler.endElementNS(("namespace", "child"), "child")

    assert len(handler._stack) == 1
    assert len(handler._stack[0]) == 1
    assert handler._stack[0][0].tag == CN.clarkFromTuple(("namespace", "child"))

def test_end_element_ns_invalid_stack(setup):
    handler = setup

    with pytest.raises(Exception):
        handler.endElementNS(("namespace", "element"), "element")


def test_end_element_ns_mismatched_tag(setup):
    handler = setup
    handler._stack.append(ET.Element("different"))

    with pytest.raises(Exception):
        handler.endElementNS(("namespace", "element"), "element")


@patch('pyjabber.stream.StanzaHandler.StanzaHandler')
@patch('pyjabber.stream.StreamHandler.StreamHandler')
@patch('pyjabber.stream.StanzaHandler.PluginManager')
@patch('pyjabber.stream.StanzaHandler.Presence')
def test_end_element_ns_stream_handling(mock_plugin_manager, mock_presence, mock_handle_open_stream, mock_stream_handler, setup):
    handler = setup
    handler._stack.append(ET.Element("{http://etherx.jabber.org/streams}stream"))

    mock_plugin_manager.return_value = MagicMock()
    mock_presence.return_value = MagicMock()

    elem = ET.Element(CN.clarkFromTuple(("namespace", "dummy")))
    handler._stack.append(elem)

    mock_handle_open_stream.return_value = Signal.DONE
    mock_stream_handler.handle_open_stream.return_value = Signal.DONE
    handler._streamHandler = mock_stream_handler

    with patch('pyjabber.network.XMLParser.StanzaHandler') as mock_stanza_handler:
        handler.endElementNS(("namespace", "dummy"), "dummy")

    assert handler._state == XMLParser.StreamState.READY


def test_characters(setup):
    handler = setup
    parent = ET.Element("parent")
    child = ET.SubElement(parent, "child")
    handler._stack.append(parent)

    # Verificación cuando el elemento tiene hijos
    handler.characters("content")
    assert child.tail == "content"

    # Verificación cuando el elemento no tiene hijos
    handler._stack[-1] = ET.Element("parent_no_children")
    handler.characters("more content")
    assert handler._stack[-1].text == "more content"
    handler.characters(" additional content")
    assert handler._stack[-1].text == "more content additional content"

if __name__ == "__main__":
    pytest.main()



from unittest.mock import patch

import pytest
import xml.etree.ElementTree as ET

from pyjabber.stanzas.error.StanzaError import (
    bad_request,
    conflict_error,
    feature_not_implemented,
    invalid_xml,
    item_not_found,
    not_acceptable,
    not_authorized,
    service_unavaliable,
    XMLNS
)


def test_bad_request():
    expected = b"<error type='modify'><bad-request xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/></error>"
    assert bad_request() == expected

def test_conflict_error():
    id = "123"
    with patch('pyjabber.stanzas.error.StanzaError.host') as mock_host:
        mock_host.get.return_value = 'localhost'
        result = conflict_error(id)

    root = ET.fromstring(result)

    assert root.tag == "iq"
    assert root.attrib["id"] == id
    assert root.attrib["type"] == "error"
    assert root.attrib["from"] == "localhost"

    error = root.find("error")
    assert error is not None
    assert error.attrib["type"] == "cancel"

    conflict = error.find(f"{{{XMLNS}}}conflict")
    if conflict is None:
        print(ET.tostring(root, encoding='unicode'))
    assert conflict is not None
    assert conflict.tag == f"{{{XMLNS}}}conflict"

    text = error.find(f"{{{XMLNS}}}text")
    assert text is not None
    assert text.tag == f"{{{XMLNS}}}text"
    assert text.text == "The requested username already exists"
def test_feature_not_implemented():
    ns = "custom:namespace"
    feature = "some_feature"
    expected = f"<error type='cancel'><feature-not-implemented xmlns='{XMLNS}'/><unsupported "f"xmlns='{ns}' feature='{feature}'/></error>".encode()
    assert feature_not_implemented(feature, ns) == expected


def test_invalid_xml():
    expected = f"<stream:error><invalid-xml xmlns='{XMLNS}'/></stream:error></stream:stream>".encode()
    assert invalid_xml() == expected


def test_item_not_found():
    expected = b"<error type='cancel'><item-not-found xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/></error>"
    assert item_not_found() == expected


def test_not_acceptable_with_text():
    text = "Error message"
    expected = f"<error type='modify'><not-acceptable xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'><text>{text}</text></not-acceptable></error>".encode()
    assert not_acceptable(text) == expected


def test_not_acceptable_without_text():
    expected = b"<error type='modify'><not-acceptable xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/></error>"
    assert not_acceptable() == expected


def test_not_authorized():
    expected = b"<failure xmlns='urn:ietf:params:xml:ns:xmpp-sasl'><not-authorized/></failure>"
    assert not_authorized() == expected


def test_service_unavaliable():
    expected = b"<error type='cancel'><service-unavailable xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/></error>"
    assert service_unavaliable() == expected


if __name__ == "__main__":
    pytest.main()



import asyncio
import os

from aiohttp import web
from loguru import logger

from pyjabber.webpage.api import api

FILE_PATH = os.path.dirname(os.path.abspath(__file__))

class AdminPage:
    def __init__(self):
        self._app = web.Application()
        self._app.router.add_get('/', self.getIndex)
        self._app.router.add_get('/api/users', api.handleUser)
        self._app.router.add_get('/api/roster/{id}', api.handleRoster)
        self._app.router.add_post('/api/createuser', api.handleRegister)
        self._app.router.add_delete('/api/users/{id}', api.handleDelete)
        self._app.router.add_static('/static', FILE_PATH + '/build/static')
        self._app.router.add_static('/', FILE_PATH + '/build/')

    @property
    def app(self):
        return self._app

    async def getIndex(self, request):
        return web.FileResponse(FILE_PATH + '/build/index.html')

    async def start(self):
        runner = web.AppRunner(self._app)
        await runner.setup()
        site = web.TCPSite(runner, 'localhost', 9090)
        try:
            await site.start()
        except OSError as e:
            logger.error(e)
            return

        logger.info("Serving admin webpage on http://localhost:9090")
        try:
            while True:
                await asyncio.sleep(3600)  # Keep alive the server
        except asyncio.CancelledError:
            await runner.cleanup()




import asyncio
import contextvars

host = contextvars.ContextVar("host", default="localhost")
config_path = contextvars.ContextVar("config_path")
cert_path = contextvars.ContextVar("cert_path")
root_path = contextvars.ContextVar("root_path")
database_path = contextvars.ContextVar("database_path")
database_in_memory = contextvars.ContextVar("database_in_memory", default=False)

tls_queue = contextvars.ContextVar("tls_queue")
connection_queue = contextvars.ContextVar("connection_queue")
message_queue = contextvars.ContextVar("message_queue")



import xml.etree.ElementTree as ET

from loguru import logger

from pyjabber import metadata
from pyjabber.features.presence.PresenceFeature import Presence
from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.stream.JID import JID
from pyjabber.plugins.PluginManager import PluginManager


class InternalServerError(Exception):
    pass


class StanzaHandler:
    def __init__(self, buffer) -> None:
        self._host = metadata.host.get()
        self._buffer = buffer
        self._connections = ConnectionManager()
        self._message_queue = metadata.message_queue.get()

        self._peername = buffer.get_extra_info('peername')
        self._jid = self._connections.get_jid(self._peername)

        self._pluginManager = PluginManager(self._jid)
        self._presenceManager = Presence()

        self._functions = {
            "{jabber:client}iq": self.handle_iq,
            "{jabber:client}message": self.handle_msg,
            "{jabber:client}presence": self.handle_pre
        }

    def feed(self, element: ET.Element):
        try:
            self._functions[element.tag](element)
        except (KeyError, InternalServerError) as e:
            logger.error(e)
            logger.error(f"INTERNAL SERVER ERROR WITH {self._peername}. CLOSING CONNECTION FOR SERVER STABILITY")
            raise InternalServerError

    def handle_iq(self, element: ET.Element):
        """
            Process the iq stanza with the PluginManager (PM) class
            If the feature/XEP requested is not available, the PM instance
            will send a

            :param element: The stanza in the ElementTree format
        """
        res = self._pluginManager.feed(element)
        if res:
            self._buffer.write(res)

    def handle_msg(self, element: ET.Element):
        """
            Router the message to the client

            If the destination client is a user of a remote server, it will queue the message into the QueueMessage
            object and try to connect to the remote server

            :param element: the message in the ElementTree format
        """
        jid = JID(element.attrib["to"])

        if jid.domain == self._host:
            if not jid.resource:
                priority = self._presenceManager.most_priority(jid)
                if not priority:
                    self._message_queue.put_nowait(('MESSAGE', jid.bare(), ET.tostring(element)))
                    return None

                all_resources_online = []
                for user in priority:
                    all_resources_online += self._connections.get_buffer_online(JID(user=jid.user, domain=jid.domain, resource=user[0]))
                for buffer in all_resources_online:
                    buffer[1].write(ET.tostring(element))
            else:
                resource_online = self._connections.get_buffer_online(jid)
                if not resource_online:
                    self._message_queue.put_nowait(('MESSAGE', str(jid), ET.tostring(element)))
                else:
                    for buffer in resource_online:
                        buffer[1].write(ET.tostring(element))

        else:
            pass
            # the s2s feature is currently disabled due to bad implementation
            # Future version of the server will fix that

            # server_buffer = self._connections.get_server_buffer(jid.bare())
            # if server_buffer:
            #     server_buffer[1].write(ET.tostring(element))
            #
            # else:
            #     self._queue_message.enqueue(jid.domain, ET.tostring(element))

    def handle_pre(self, element: ET.Element):
        """
            Handle the presences stanzas
        """
        res = self._presenceManager.feed(self._jid, element)
        if res:
            self._buffer.write(res)



from asyncio import Transport
from contextlib import closing
from itertools import chain
from typing import List, Tuple
from uuid import uuid4
from xml.etree import ElementTree as ET

from loguru import logger
from yaml import load, Loader

from pyjabber.metadata import host, config_path
from pyjabber.db.database import connection
from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.plugins.xep_0060.enum import NodeAttrib, SubscribersAttrib, Subscription, Affiliation
from pyjabber.plugins.xep_0060.error import ErrorType
from pyjabber.plugins.xep_0060.error import error_response
from pyjabber.stanzas.IQ import IQ
from pyjabber.stanzas.Message import Message
from pyjabber.stanzas.error import StanzaError
from pyjabber.stream.JID import JID
from pyjabber.utils import Singleton, ClarkNotation as CN


def success_response(element: ET.Element, owner: bool = False) -> Tuple[IQ, ET.Element]:
    iq_res = IQ(
        type_=IQ.TYPE.RESULT,
        from_=host.get(),
        to=element.attrib.get('from'),
        id_=element.attrib.get('id') or str(uuid4())
    )
    if owner:
        xmlns = 'http://jabber.org/protocol/pubsub#owner'
    else:
        xmlns = 'http://jabber.org/protocol/pubsub'

    pubsub = ET.SubElement(iq_res, 'pubsub', attrib={'xmlns': xmlns})
    return iq_res, pubsub


class PubSub(metaclass=Singleton):
    def __init__(self, db_connection_factory=None):
        super().__init__()

        self._connections = ConnectionManager()

        items = load(open(config_path.get()), Loader=Loader)['items']
        service_jid = next((s for s in list(items) if 'pubsub' in s), None)
        if service_jid is None:
            raise Exception  # TODO: Define missing config exception

        self._jid = service_jid
        self._category = items.get('type')
        self._ver = items.get('var')

        self._host = host.get()
        self._db_connection_factory = db_connection_factory or connection

        self._nodes = None
        self._subscribers = None
        self.update_memory_from_database()

        self._operations = {
            'create': self.create_node,
            'delete': self.delete_node,
            'subscribe': self.subscribe,
            'unsubscribe': self.unsubscribe,
            'subscriptions': self.retrieve_subscriptions,
            'publish': self.publish
        }

    def update_memory_from_database(self):
        with closing(self._db_connection_factory()) as con:
            res = con.execute("SELECT * FROM pubsub")
            self._nodes = res.fetchall()
            res = con.execute("SELECT * FROM pubsubSubscribers")
            self._subscribers = res.fetchall()

    def feed(self, jid: JID, element: ET.Element):
        try:
            _, tag = CN.deglose(element[0].tag)

            if tag != 'pubsub':
                return StanzaError.invalid_xml()

            _, operation = CN.deglose(element[0][0].tag)
            return self._operations[operation](element, jid)
        except (KeyError, TypeError) as e:
            logger.error(e)  # TODO: Malformed request

    def discover_items(self, element: ET.Element) -> List[tuple]:
        """
        Returns the available nodes at the level specified in the query
        :return: A list of 3-tuples in the format (node, name, type)
        """
        query = element.find('{http://jabber.org/protocol/disco#items}query')
        if query is not None and query.attrib.get('node') is None:  # Query to root
            res = []
            for node in self._nodes:
                _node = node[NodeAttrib.NODE.value]
                _name = node[NodeAttrib.NAME.value]
                _type = node[NodeAttrib.TYPE.value]
                res.append((_node, _name, _type))
            return res

        else:  # Query to branch/leaf in the nodes tree
            pass

    def discover_info(self, element: ET.Element):
        """
        Return the info for a given node
        :return: A 2-tuple in the format of (name, type)
        """
        query = element.find('{http://jabber.org/protocol/disco#info}query')
        query_node = query.attrib.get('node')
        if query is not None and query_node is not None:
            match_node = next((node for node in self._nodes if node[NodeAttrib.NODE] == query_node), None)
            if match_node:
                return match_node[NodeAttrib.NAME], match_node[NodeAttrib.TYPE]

        return None

    def create_node(self, element: ET.Element, jid: JID):
        """
        Creates a new node in the pubsub service.
        The owner will be the creator of the node.
        Configuration can be included in the creation with a form in the IQ request.
        """
        pubsub = element.find('{http://jabber.org/protocol/pubsub}pubsub')
        create = pubsub.find('{http://jabber.org/protocol/pubsub}create')
        config = pubsub.find('{http://jabber.org/protocol/pubsub}config')

        new_node = create.attrib.get('node')
        if new_node is None:
            return error_response(element, jid, ErrorType.NOT_ACCEPTABLE)

        # Node already exists
        if [node for node in self._nodes if node[NodeAttrib.NODE.value] == new_node]:
            return error_response(element, jid, ErrorType.CONFLICT)

        if config:
            pass  # TODO: create node with given configuration

        """
        A new item MUST follow the order described in the NodeAttrib enum
        for correct attribute access
        """
        item = (
            new_node,  # NODE
            jid.user,  # OWNER
            None,  # NAME
            'leaf',  # TYPE
            1024  # MAX ITEMS
        )

        with closing(self._db_connection_factory()) as con:
            con.execute("INSERT INTO pubsub VALUES (?,?,?,?,?)", item)
            con.commit()

        self.update_memory_from_database()

        iq_res, pubsub = success_response(element)
        ET.SubElement(pubsub, 'create', attrib={'node': new_node})
        return ET.tostring(iq_res)

    def delete_node(self, element: ET.Element, jid: JID):
        """
        Deletes a specific node in the pubsub service.
        ONLY the owner has the permissions to delete.
        """
        del_node = element[0][0].attrib.get('node')
        if del_node is None:
            return error_response(element, jid, ErrorType.NOT_ACCEPTABLE)

        try:
            node_match = [node for node in self._nodes if node[NodeAttrib.NODE.value] == del_node][0]
        except IndexError:
            return error_response(element, jid, ErrorType.ITEM_NOT_FOUND)

        if node_match[NodeAttrib.OWNER.value] != jid.user:
            return error_response(element, jid, ErrorType.FORBIDDEN)

        with closing(self._db_connection_factory()) as con:
            con.execute("DELETE FROM pubsub WHERE node = ?", (del_node,))
            con.commit()

        self.update_memory_from_database()
        _, suc_res = success_response(element)
        return ET.tostring(suc_res)

    def subscribe(self, element: ET.Element, jid: JID):
        """
        Subscribe to a specific node
        The default affiliation will be MEMBER
        """
        pubsub = element.find('{http://jabber.org/protocol/pubsub}pubsub')
        subscribe = pubsub.find('{http://jabber.org/protocol/pubsub}subscribe')
        node = subscribe.attrib.get('node')
        jid_request = subscribe.attrib.get('jid')

        if node is None:
            return error_response(element, jid, ErrorType.NOT_ACCEPTABLE)

        if jid_request is None:
            return error_response(element, jid, ErrorType.INVALID_JID)

        jid_request = JID(jid_request)
        if jid_request.bare() != jid.bare():
            return error_response(element, jid, ErrorType.INVALID_JID)

        try:
            target_node = [n for n in self._nodes if n[NodeAttrib.NODE.value] == node][0]
        except IndexError:
            return error_response(element, jid, ErrorType.ITEM_NOT_FOUND)

        current_state = [
            s for s in self._subscribers
            if s[SubscribersAttrib.JID.value] == jid_request.user and s[SubscribersAttrib.NODE.value] == node
        ]
        if len(current_state) >= 1:
            current_state = current_state[0]
            if (current_state[SubscribersAttrib.SUBSCRIPTION.value]
               in [Subscription.SUBSCRIBED.value, Subscription.UNCONFIGURED.value]):

                iq_res, pubsub = success_response(element)
                subid = current_state[SubscribersAttrib.SUBID.value]
                ET.SubElement(
                    pubsub,
                    'subscription',
                    attrib={
                        'node': target_node[NodeAttrib.NODE.value],
                        'jid': jid_request.bare(),
                        'subid': subid,
                        'subscription': 'subscribed'
                    }
                )
                return ET.tostring(iq_res)

            elif current_state[SubscribersAttrib.SUBSCRIPTION.value] == Subscription.PENDING.value:
                return error_response(element, jid, ErrorType.PENDING_SUBSCRIPTION)

        subid = str(uuid4())

        item = (
            target_node[NodeAttrib.NODE.value],
            jid_request.user,
            subid,
            Subscription.SUBSCRIBED.value,
            Affiliation.MEMBER
        )

        with closing(self._db_connection_factory()) as con:
            con.execute("INSERT INTO pubsubSubscribers VALUES (?,?,?,?,?)", item)
            con.commit()

        self.update_memory_from_database()

        iq_res, pubsub = success_response(element)
        ET.SubElement(
            pubsub,
            'subscription',
            attrib={
                'node': target_node[NodeAttrib.NODE.value],
                'jid': jid_request.bare(),
                'subid': subid,
                'subscription': 'subscribed'
            }
        )
        return ET.tostring(iq_res)

    def unsubscribe(self, element: ET.Element, jid: JID):
        """
        Unsubscribe to a specific node
        """
        pubsub = element.find('{http://jabber.org/protocol/pubsub}pubsub')
        unsubscribe = pubsub.find('{http://jabber.org/protocol/pubsub}unsubscribe')
        node = unsubscribe.attrib.get('node')
        jid_request = unsubscribe.attrib.get('jid')
        subid = unsubscribe.attrib.get('subid')

        if node is None:
            return error_response(element, jid, ErrorType.NOT_ACCEPTABLE)

        if jid_request is None:
            return error_response(element, jid, ErrorType.INVALID_JID)

        jid_request = JID(jid_request)
        if jid_request.bare() != jid.bare():
            return error_response(element, jid, ErrorType.INVALID_JID)

        try:
            target_node = [n for n in self._nodes if n[NodeAttrib.NODE.value] == node][0]
        except KeyError:
            return error_response(element, jid, ErrorType.ITEM_NOT_FOUND)

        current_state = [s for s in self._subscribers if s[SubscribersAttrib.JID.value] == jid_request.user]

        if len(current_state) == 0:
            return error_response(element, jid, ErrorType.NOT_SUBSCRIBED)

        if len(current_state) > 1:
            if subid is None:
                return error_response(element, jid, ErrorType.SUBID_REQUIRED)

            if len([s for s in self._subscribers if s[SubscribersAttrib.SUBID.value] == subid]) == 0:
                return error_response(element, jid, ErrorType.INVALID_SUBID)

            query = "DELETE FROM pubsubSubscribers WHERE node = ? AND jid = ? AND subid = ?"
            item = (target_node[NodeAttrib.NODE.value], jid_request.user, subid)

        else:
            query = "DELETE FROM pubsubSubscribers WHERE node = ? AND jid = ?"
            item = (target_node[NodeAttrib.NODE.value], jid_request.user)

        with closing(self._db_connection_factory()) as con:
            con.execute(query, item)
            con.commit()

        self.update_memory_from_database()

        iq_res, pubsub = success_response(element)
        sub = ET.SubElement(
            pubsub,
            'subscription',
            attrib={
                'node': target_node[NodeAttrib.NODE.value],
                'jid': jid_request.bare(),
                'subscription': 'none'
            }
        )
        if subid: sub.attrib['subid'] = subid
        return ET.tostring(iq_res)

    def retrieve_subscriptions(self, element: ET.Element, jid: JID):
        pubsub = element.find('{http://jabber.org/protocol/pubsub}pubsub') or element.find(
            '{http://jabber.org/protocol/pubsub#owner}pubsub')
        subscriptions = pubsub.find('{http://jabber.org/protocol/pubsub}subscriptions') or pubsub.find(
            '{http://jabber.org/protocol/pubsub#owner}subscriptions')
        target_node = subscriptions.attrib.get('node')
        from_stanza = element.attrib.get('from')

        if from_stanza is not None and JID(from_stanza).user != jid.user:
            return error_response(element, jid, ErrorType.FORBIDDEN)

        iq_res, pubsub = success_response(element)
        # pubsub_res = ET.SubElement(iq_res, 'pubsub' ,attrib={'xmlns': 'http://jabber.org/protocol/pubsub'})
        subscriptions_res = ET.SubElement(pubsub, 'subscriptions')

        if target_node is not None and target_node != '':
            query = "SELECT node, subscription, subid FROM pubsubSubscribers WHERE jid = ? AND node = ?"
            item = (str(jid.user), target_node)
        else:
            query = "SELECT node, subscription, subid FROM pubsubSubscribers WHERE jid = ?"
            item = (str(jid.user),)

        with closing(self._db_connection_factory()) as con:
            res = con.execute(query, item)
            res = res.fetchall()

        for sub in res:
            ET.SubElement(subscriptions_res, 'subscription', attrib={
                'node': sub[0],
                'jid': jid.bare(),
                'subscription': sub[1],
                'subid': sub[2]
            })

        return ET.tostring(iq_res)

    def retrieve_affiliations(self, element: ET.Element, jid: str):
        pass

    def publish(self, element: ET.Element, jid: JID):
        pubsub = element.find('{http://jabber.org/protocol/pubsub}pubsub')
        publish = pubsub.find('{http://jabber.org/protocol/pubsub}publish')
        item = publish.find('{http://jabber.org/protocol/pubsub}item')
        payload = item[0]
        node = publish.attrib.get('node')

        if len(item) > 1:
            return error_response(element, jid, ErrorType.INVALID_PAYLOAD)

        target_node = [n for n in self._nodes if n[NodeAttrib.NODE.value] == node]
        if len(target_node) == 0:
            return error_response(element, jid, ErrorType.ITEM_NOT_FOUND)

        current_sub = [s for s in self._subscribers if s[SubscribersAttrib.AFFILIATION.value] == jid.user]
        if jid.user != target_node[0][NodeAttrib.OWNER.value] \
            or (current_sub and current_sub[0][SubscribersAttrib.AFFILIATION.value] != Affiliation.MEMBER):
            return error_response(element, jid, ErrorType.FORBIDDEN)

        item_id = item.attrib.get('id')

        with closing(self._db_connection_factory()) as con:
            if item_id is not None:
                res = con.execute("SELECT itemid FROM pubsubItems WHERE itemid = ?", item_id)
                if res.fetchone():
                    con.execute("UPDATE pubsubItems "
                                "SET payload = ? "
                                "WHERE node = ? AND itemid = ?",
                                (target_node[NodeAttrib.NODE.value], item_id, ET.tostring(payload)))

            else:
                item_id = str(uuid4())
                con.execute("INSERT INTO pubsubItems VALUES (?,?,?)",
                            (target_node[0][NodeAttrib.NODE.value], item_id, ET.tostring(payload)))
                con.commit()

        self.send_notification(target_node[0][NodeAttrib.NODE.value], payload)

        iq_res, pubsub = success_response(element)
        publish = ET.SubElement(pubsub, 'publish', attrib={'node': 'node'})
        if item_id:
            ET.SubElement(publish, 'item', attrib={'id': item_id})
        return ET.tostring(iq_res)

    def send_notification(self, node: str, payload: ET.Element, item_id: str = None):
        receivers = [s for s in self._subscribers
                     if s[SubscribersAttrib.NODE.value] == node
                     and s[SubscribersAttrib.AFFILIATION.value] == Affiliation.MEMBER]

        receivers_jid = [r[1] for r in receivers]
        receivers_buffer = [self._connections.get_buffer(JID(user=r, domain=self._host)) for r in receivers_jid]
        receivers_buffer_single_iterator: List[Tuple[JID, Transport]] = list(chain.from_iterable(receivers_buffer))

        event = ET.Element('event', attrib={'xmlns': 'http://jabber.org/protocol/pubsub#event'})
        items = ET.SubElement(event, 'item')
        if item_id: items.attrib['id'] = item_id
        items.append(payload)

        for jid, buffer in receivers_buffer_single_iterator:
            message = Message(
                mto=jid.bare(),
                mfrom=self._host,
                id=str(uuid4()),
                mtype=None,
                body=event
            )
            buffer.write(ET.tostring(message))



import asyncio
import os

from loguru import logger
from xml import sax
from xml.etree.ElementTree import Element

from pyjabber import metadata
from pyjabber.features.presence.PresenceFeature import Presence
from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.network.StreamAlivenessMonitor import StreamAlivenessMonitor
from pyjabber.network.XMLParser import XMLParser
from pyjabber.stream.StanzaHandler import InternalServerError

FILE_AUTH = os.path.dirname(os.path.abspath(__file__))


class TransportProxy:
    def __init__(self, transport, peer):
        self._transport = transport
        self._peer = peer

    @property
    def originalTransport(self):
        return self._transport

    def write(self, data):
        logger.trace(f"Sending to {self._peer}: {data}")
        return self._transport.write(data)

    def __getattr__(self, name):
        return getattr(self._transport, name)


class XMLProtocol(asyncio.Protocol):
    """
    Protocol to manage the network connection between nodes in the XMPP network. Handles the transport layer.

    :param namespace: namespace of the XML tags (jabber:client or jabber:server)
    :param host: Host for connections
    :param connection_timeout: Max time without any response from a client. After that, the server will terminate the connection
    :param cert_path: Path to custom domain certs. By default, the server generates its own certificates for hostname
    """

    def __init__(
            self,
            namespace,
            host,
            connection_timeout,
            cert_path):

        self._xmlns = namespace
        self._host = host
        self._connection_timeout = connection_timeout
        self._connection_manager = ConnectionManager()
        self._presence_manager = Presence()
        self._cert_path = cert_path
        self._tls_queue = metadata.tls_queue.get()

        self._transport = None
        self._peer = None
        self._xml_parser = None
        self._timeout_monitor = None
        self._timeout_flag = False


    def connection_made(self, transport):
        """
        Called when a client or another server opens a TCP connection to the server

        :param transport: The transport object for the connection
        :type transport: asyncio.Transport
        """
        if transport:
            self._peer = transport.get_extra_info('peername')
            self._transport = TransportProxy(transport, self._peer)

            self._xml_parser = sax.make_parser()
            self._xml_parser.setFeature(sax.handler.feature_namespaces, True)
            self._xml_parser.setFeature(sax.handler.feature_external_ges, False)
            self._xml_parser.setContentHandler(
                XMLParser(self._transport, self.task_tls)
            )

            if self._connection_timeout:
                self._timeout_monitor = StreamAlivenessMonitor(
                    timeout=self._connection_timeout,
                    callback=self.connection_timeout
                )

            self._connection_manager.connection(self._peer, self._transport)

            logger.info(f"Connection from {self._peer}")
        else:
            logger.error("Invalid transport")

    def connection_lost(self, exc):
        """
        Called when a client or another server closes a TCP connection to the server

        :param exc: Exception that caused the connection to close
        :type exc: Exception
        """
        if self._timeout_flag:
            return

        logger.info(f"Connection lost from {self._peer}: Reason {exc}")
        jid = self._connection_manager.get_jid(self._peer)
        if jid and jid.user and jid.domain:
            self._presence_manager.feed(jid, Element("presence", attrib={"type": "INTERNAL"}))

        self._transport = None
        self._xml_parser = None
        self._timeout_monitor.cancel()

        self._connection_manager.disconnection(self._peer)

    def data_received(self, data):
        """
        Called when data is received from the client or another server

        :param data: Chunk of data received
        :type data: Byte array
        """
        try:
            logger.debug(f"Data received from <{self._peer}>: {data.decode()}")
        except:
            logger.debug(f"Binary data recived")

        if self._timeout_monitor:
            self._timeout_monitor.reset()

        '''
        Some XMPP clients/libraries send the XML header
        with the stream stanza all together.
        This can lead sometimes in a exception for the parser.
        An XML header gives no usefull information, as the RFC6120
        takes the vesion 1.0 as standar, so we can remove it safely.
        If I try to remove the header inside the except block, the parser
        keeps refusing it. I'm forced to do always before the feed.
        I probably should change the parser
        '''
        data = data.replace(b"<?xml version=\'1.0\'?>", b"")
        data = data.replace(b"<?xml version=\"1.0\"?>", b"")

        try:
            self._xml_parser.feed(data)
        except InternalServerError:
            a = 1
            pass

    def eof_received(self):
        """
        Called when the client or another server sends an EOF
        """
        if self._transport:
            logger.debug(f"EOF received from {self._peer}")

    def connection_timeout(self):
        """
        Called when the stream is not responding for a long tikem
        """
        logger.info(f"Connection timeout from {self._peer}")

        try:
            self._transport.write("<connection-timeout/>".encode())
            self._transport.close()
        except:
            logger.warning(f"Connection with {self._peer} is already closed. Removing from online list")

        self._connection_manager.disconnection(self._peer)

        self._transport = None
        self._xml_parser = None
        self._timeout_flag = True

    def task_tls(self):
        """
            Sync function to call the STARTTLS coroutine
        """
        self._tls_queue.put_nowait((self._transport, self, self._xml_parser.getContentHandler()))



from enum import Enum
from typing import List

from xml.etree import ElementTree as ET

from pyjabber.plugins.xep_0004.field import FieldRequest, FieldResponse, FieldTypes
from pyjabber.utils import ClarkNotation as CN
from pyjabber.stanzas.error import StanzaError as SE


class MissingDataForms(Exception):
    """
    No dataforms founded in a stanza, when it was expected
    """
    pass


class FormType(Enum):
    FORM = 'form'
    SUBMIT = 'submit'
    CANCEL = 'cancel'
    RESULT = 'result'


def parse_form(element: ET.Element):
    """
    If a proper dataforms is passes, it returns a list of field
    submitted by the client
    @param element:
    @return: List of dicts in format {type, var, values[]}
    """
    data = None
    try:
        """
        Supposing element is an IQ stanza, it should be in a second-level child
        """
        forms = element[0][0]
        ns, tag = CN.deglose(forms.tag)
        if tag != 'x' or ns != 'jabber:x:data':
            raise MissingDataForms
        data = forms
    except (KeyError, MissingDataForms):
        """
        Supposing element is an Message stanza, it should be in a first-level child
        """
        try:
            forms = element[0]
            ns, tag = CN.deglose(forms.tag)
            if tag != 'x' or ns != 'jabber:x:data':
                raise MissingDataForms
            data = forms
        except (KeyError, MissingDataForms):
            return SE.bad_request()

    field_list = data.findall('{jabber:x:data}field')
    try:
        field_list = [
            FieldResponse(
                field_type=FieldTypes.from_value(f.attrib.get('type')),
                var=f.attrib.get('var'),
                values=[v.text for v in f.findall('{jabber:x:data}value')]
            ) for f in field_list]
    except KeyError:
        return SE.bad_request()

    return field_list


def generate_form(form_type: FormType, title: str = None, instructions: str = None, fields: List[FieldRequest] = None) -> ET.Element:
    form_res = ET.Element('x', attrib={'xmlns': 'jabber:x:data', 'type': form_type.value})

    if form_type == FormType.CANCEL.value:
        return form_res

    if title:
        ET.SubElement(form_res, 'title').text = title

    if instructions:
        ET.SubElement(form_res, 'instructions').text = instructions

    for f in fields:
        field = ET.Element('field', attrib={
            'type': f.type.value,
            'var': f.var
        })

        if f.label:
            field.attrib['label'] = f.label

        for v in f.values:
            value = ET.Element('value')
            value.text = v
            field.append(value)

        if f.options:
            for o in f.options:
                option = ET.Element('option')
                option.text = o
                field.append(option)

        if f.desc:
            desc = ET.Element('desc')
            desc.text = f.desc
            field.append(desc)

        form_res.append(field)

    return form_res


def generate_form_multi_res():
    pass



from enum import Enum
from typing import Union
from uuid import uuid4
from xml.etree import ElementTree as ET

from pyjabber.features import InBandRegistration as IBR
from pyjabber.features.StartTLSFeature import StartTLSFeature, proceed_response
from pyjabber.features.StreamFeature import StreamFeature
from pyjabber.features.SASLFeature import SASLFeature, SASL
from pyjabber.features.ResourceBinding import ResourceBinding
from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.metadata import host
from pyjabber.stanzas.IQ import IQ


class Stage(Enum):
    """
    Stream connection states.
    """
    CONNECTED = 0
    OPENED = 1
    SSL = 2
    SASL = 3
    AUTH = 4
    BIND = 5
    READY = 6


class Signal(Enum):
    RESET = 0
    DONE = 1

    def __eq__(self, other):
        if other is None:
            return False
        return self.value == other.value


class StreamHandler:
    def __init__(self, buffer, starttls) -> None: # connection_manager
        self._host = host.get()
        self._buffer = buffer
        self._starttls = starttls

        self._streamFeature = StreamFeature()
        self._connection_manager: ConnectionManager = ConnectionManager()
        self._stage = Stage.CONNECTED

        self._elem = None
        self._jid = None

    @property
    def buffer(self):
        return self._buffer

    @buffer.setter
    def buffer(self, value):
        self._buffer = value

    def handle_open_stream(self, elem: ET.Element = None) -> Union[Signal, None]:
        # TCP connection opened. The server returns the available features (only TLS)
        if self._stage == Stage.CONNECTED:
            self._streamFeature.reset()
            self._streamFeature.register(StartTLSFeature())
            self._buffer.write(self._streamFeature.to_bytes())

            self._stage = Stage.OPENED
            return

        # TLS offered. The client should have responded with a starttls message
        elif self._stage == Stage.OPENED:
            if "starttls" in elem.tag:
                self._buffer.write(proceed_response())
                self._starttls()
                self._stage = Stage.SSL
                self._buffer.pause_reading()
                return Signal.RESET

            else:
                raise Exception()

        # TLS Handshake made. Starting SASL
        elif self._stage == Stage.SSL:
            self._streamFeature.reset()

            self._streamFeature.register(IBR.InBandRegistration())
            self._streamFeature.register(SASLFeature())
            self._buffer.write(self._streamFeature.to_bytes())

            self._stage = Stage.SASL

        # SASL
        elif self._stage == Stage.SASL:
            res = SASL().feed(elem, {"peername": self._buffer.get_extra_info('peername')})
            if type(res) is tuple:
                if res[0].value == Signal.RESET.value:
                    self._buffer.write(res[1])
                    self._stage = Stage.AUTH
                    return Signal.RESET
                else:
                    self._buffer.write(res[1])
                    self._stage = Stage.AUTH
                    return

            self._buffer.write(res)

        # User register/authenticated. Starting resource binding
        elif self._stage == Stage.AUTH:
            self._streamFeature.reset()
            self._streamFeature.register(ResourceBinding())
            self._buffer.write(self._streamFeature.to_bytes())

            self._stage = Stage.BIND

        elif self._stage == Stage.BIND:
            if "iq" in elem.tag:
                if elem.attrib["type"] == "set":
                    resource_id = str(uuid4())

                    iq_res = IQ(type_=IQ.TYPE.RESULT, id_=elem.get('id') or str(uuid4()))
                    bind_res = ET.SubElement(iq_res, "bind", attrib={"xmlns": "urn:ietf:params:xml:ns:xmpp-bind"})

                    peername = self._buffer.get_extra_info('peername')
                    new_jid = self._connection_manager.get_jid(peername)
                    new_jid.resource = resource_id

                    ET.SubElement(bind_res, 'jid').text = str(new_jid)

                    self._buffer.write(ET.tostring(iq_res))

                    """
                    Stream is negotiated.
                    Update the connection register with the jid and transport
                    """
                    self._connection_manager.set_jid(peername, new_jid, self._buffer)

            return Signal.DONE



import os
from xml import sax

from loguru import logger

from pyjabber.network.server.incoming.XMLServerIncomingParser import (
    XMLServerIncomingParser,
)
from pyjabber.network.StreamAlivenessMonitor import StreamAlivenessMonitor
from pyjabber.network.XMLProtocol import XMLProtocol

FILE_AUTH = os.path.dirname(os.path.abspath(__file__))


class XMLServerIncomingProtocol(XMLProtocol):
    """
    Protocol to manage the network connection between nodes in the XMPP network. Handles the transport layer.
    """

    def __init__(
            self,
            namespace,
            host,
            connection_timeout,
            cert_path):

        super().__init__(
            namespace,
            host,
            connection_timeout,
            cert_path)

    def connection_made(self, transport):
        """
        Called when a client or another server opens a TCP connection to the server

        :param transport: The transport object for the connection
        :type transport: asyncio.Transport
        """
        if transport:
            self._transport = transport

            self._xml_parser = sax.make_parser()
            self._xml_parser.setFeature(sax.handler.feature_namespaces, True)
            self._xml_parser.setFeature(
                sax.handler.feature_external_ges, False)
            self._xml_parser.setContentHandler(
                XMLServerIncomingParser(
                    self._host,
                    self._transport,
                    self.task_tls
                )
            )

            if self._connection_timeout:
                self._timeout_monitor = StreamAlivenessMonitor(
                    timeout=self._connection_timeout,
                    callback=self.connection_timeout
                )

            self._connection_manager.connection(
                self._transport.get_extra_info('peername'))

            logger.info(
                f"Server connection to {self._transport.get_extra_info('peername')}")

        else:
            logger.error("Invalid transport")

    def eof_received(self):
        """
        Called when the client or another server sends an EOF
        """
        peer = self._transport.get_extra_info('peername')

        logger.debug(f"EOF received from {peer}")

        self._connection_manager.disconnection_server(peer)

        self._transport = None
        self._xml_parser = None



import asyncio
import re
from typing import List, Optional
from asyncio import Transport
from typing import Dict, Union, Tuple
from loguru import logger

from pyjabber.stream.JID import JID
from pyjabber.utils import Singleton

PeerList = Dict[
    Tuple[str, int],
    Tuple[Optional[JID], Transport, List[bool]]
]


# EXAMPLE
# peerKeys = {
#     ("0.0.0.0", "50000"): ("demo@localhost/12341234", <asyncio.Transport>, <Online: Bool>),
#     ("0.0.0.0", "50001"): ("test@localhos/43214321", <asyncio.Transport>, <Online: Bool>)
# }


class ConnectionManager(metaclass=Singleton):
    def __init__(self) -> None:

        self._peerList: PeerList = {}
        self._remoteList = {}

    ###################
    ### LOCAL BOUND ###
    ###################

    def connection(self, peer: Tuple[str, int], transport: Transport = None) -> None:
        """
            Store a new connection, without jid or transport.
            Those will be added in the future with the set_jid method.

            :param peer: The peer value in the tuple format ({IP}, {PORT})
            :param transport: The transport object associated to the connection
        """
        if peer not in self._peerList:
            self._peerList[peer] = (None, transport, [False])

    def disconnection(self, peer: Tuple[str, int]) -> None:
        """
            Remove a stored connection, and fires the DisconnectEvent

            :param peer: The peer value in the tuple format ({IP}, {PORT})
        """

        try:
            self._peerList.pop(peer)
        except KeyError:
            logger.warning(f"{peer} not present in the online list")

    def online(self, jid: JID, online: bool = True):
        for k, v in self._peerList.items():
            if v[0] == jid and v[2][0] != online:
                self._peerList[k] = (v[0], v[1], [online])

    def close(self, peer: Tuple[str, int]) -> None:
        """
            Closes a connection by sending a '</stream:stream> message' and
            deletes it from the peers list

            :param peer: The peer value in the tuple format ({IP}, {PORT})
        """
        try:
            _, buffer, _ = self._peerList.pop(peer)
            buffer.write('</stream:stream>'.encode())
            self.disconnection(peer)
        except KeyError as e:
            logger.error(f"{peer} not present in the online list")

    def get_users_connected(self) -> PeerList:
        """
            Return all the users connected

            :return: peerList
        """
        return self._peerList

    def get_buffer(self, jid: JID) -> List[Tuple[JID, Transport, bool]]:
        """
            Get all the available buffers associated with a jid.

            If the jid is in the format username@domain/resource, the function
            will only return one buffer or none.

            If the jid is in the format username@domain, the fun<Online: Bool>ction
            will return a list of the buffers for each resource available.

            Both cases return a list.

            :param jid: The jid to get the buffers for. It can be a full jid or a bare jid
            :return: (JID, TRANSPORT) tuple
        """
        if jid.domain is None:
            logger.error("JID must have, at least, user and domain")
            return []

        if jid.resource:
            return [(jid_stored, buffer, online[0])
                    for jid_stored, buffer, online in self._peerList.values()
                    if str(jid) == jid_stored]
        else:
            return [(jid_stored, buffer, online[0])
                    for jid_stored, buffer, online in self._peerList.values()
                    if re.match(f"{str(jid)}/*", str(jid_stored))]

    def get_buffer_online(self, jid: JID) -> List[Tuple[JID, Transport, bool]]:
        """
            Get all the available buffers associated with a jid
            that are ready to receive messages (online).

            - If the jid is in the format username@domain/resource, the function will only return one buffer or none.

            - If the jid is in the format username@domain, the function will return a list of the buffers for each resource available.

            Both cases return a list.

            :param jid: The jid to get the buffers for. It can be a full jid or a bare jid
            :return: (JID, TRANSPORT) tuple
        """
        if jid.domain is None:
            logger.error("JID must have, at least, user and domain")
            return []

        if jid.resource:
            return [(jid_stored, buffer, online[0])
                    for jid_stored, buffer, online in self._peerList.values()
                    if jid == jid_stored and online[0] is True]
        else:
            return [(jid_stored, buffer, online[0])
                    for jid_stored, buffer, online in self._peerList.values()
                    if re.match(f"{str(jid)}/*", str(jid_stored))
                    and online[0] is True]

    ###########
    ### JID ###
    ###########

    def get_jid(self, peer: Tuple[str, int]) -> Union[JID, None]:
        """
            Return the jid associated with the peername

            :param peer: The peer value in the tuple format ({IP}, {PORT})
        """
        try:
            return self._peerList[peer][0]
        except KeyError:
            return None

    def set_jid(self, peer: Tuple[str, int], jid: JID, transport: Transport = None) -> None:
        """
            Set/update the jid of a registered connection.

            An optional transport argument can be provided, in order to
            set/update the stored buffer

            :param peer: The peer value in the tuple format ({IP}, {PORT})
            :param jid: The jid to set/update
            :param transport: Transport to use
        """
        try:
            _, old_transport, online = self._peerList[peer]
            self._peerList[peer] = (jid, transport or old_transport, online)
        except KeyError:
            logger.error(f"Unable to find {peer} during jid/transport update")

    def update_resource(self, peer: Tuple[str, int], resource: str):
        try:
            self._peerList[peer][0].resource = resource
        except KeyError:
            logger.warning(f"Unable to find {peer} during resource update")

    ###########################################################################
    ############################# REMOTE SERVER ###############################
    ###########################################################################

    def connection_server(self, peer, host, transport) -> None:
        """
            Store a new connection, without jid or transport.
        """
        if peer not in self._remoteList:
            self._remoteList[peer] = {
                peerKeys.JID: host,
                peerKeys.TRANSPORT: None
            }

    def disconnection_server(self, peer) -> None:
        """
        Remove a present server connection in the list
        i.e. EOF recived, or TCP connection lost
        """
        try:
            self._remoteList.pop(peer)
        except KeyError:
            logger.error(f"{peer} not present in the online list")

    def get_server_buffer(self, host: str) -> Union[Tuple[str, Transport], None]:
        """
            Return the buffer associated with the given host
            :return: (<HOST>, <TRANSPORT>) tuple
        """
        if self.check_server_stream_available(host):
            key = next((k for k, v in self._remoteList.items() if v.get(peerKeys.JID) == host), None)
            return self._remoteList[key][peerKeys.JID], self._remoteList[key][peerKeys.TRANSPORT]

        if self.check_server_present_in_list(host):
            return

        self._task_s2s(host)

    def get_server_host(self, peer: Tuple[str, int]):
        """
            Return the host associated with the given peer.
            :return: Hostname
        """
        try:
            return self._remoteList[peer]["host"]
        except KeyError:
            return None

    def set_server_transport(self, peer: Tuple[str, int], transport: Transport) -> Union[None, bool]:
        """
            Set/update the transport of a registered server connection.

            :param peer: The peer value in the tuple format ({IP}, {PORT})
            :param transport: Transport to use
        """
        try:
            self._remoteList[peer][peerKeys.TRANSPORT] = transport
        except KeyError:
            return False

    def set_server_host(self, peer: Tuple[str, int], host: str):
        """
            Set/update the host of a registered server connection.

            :param peer: The peer value in the tuple format ({IP}, {PORT})
            :param host: New host name
        """
        try:
            self._remoteList[peer]["host"] = host
        except KeyError:
            return False

    def check_server_stream_available(self, host) -> bool:
        for k, v in self._remoteList.items():
            if v[peerKeys.JID] == host:
                return v[peerKeys.TRANSPORT] is not None

    def check_server_present_in_list(self, host) -> bool:
        try:
            if [v for v in self._remoteList.values() if v[self.JID] == host]:
                return True
        except KeyError:
            return False



from enum import Enum


class NodeAttrib(Enum):
    NODE = 0
    OWNER = 1
    NAME = 2
    TYPE = 3
    MAXITEMS = 4


class SubscribersAttrib(Enum):
    NODE = 0
    JID = 1
    SUBID = 2
    SUBSCRIPTION = 3
    AFFILIATION = 4


class Subscription(Enum):
    NONE = 'none'
    PENDING = 'pending'
    UNCONFIGURED = 'unconfigured'
    SUBSCRIBED = 'subscribed'


class NodeAccess(Enum):
    OPEN = 0
    PRESENCE = 1
    ROSTER = 2
    AUTHORIZE = 3
    WHITELIST = 4


class Affiliation:
    OWNER = 'owner'
    PUBLISHER = 'publisher'
    MEMBER = 'member'
    NONE = 'none'
    OUTCAST = 'outcast'



from typing import Dict
from xml.etree import ElementTree as ET


class NonUniqueFeature(Exception):
    """
    Exception raised when a non-unique feature is found on the manager
    """
    pass


class StreamFeature(ET.Element):
    """
    Class to manage server features
    """
    def __init__(
        self,
        tag: str = "stream:features",
            **extra: str) -> None:

        attrib: Dict[str, str] = {
            "xmlns": "http://etherx.jabber.org/streams"
        }

        super().__init__(tag, attrib, **extra)
        self._features: Dict[str, ET.Element] = {}

    def register(self, feature: ET.Element):
        """
        Register a new feature
        """
        if feature.tag in self._features.keys():
            raise NonUniqueFeature("Feature already registered")
        self._features[feature.tag] = feature

    def unregister(self, feature):
        """
        Unregister a feature
        """
        if feature.tag in self._features:
            del self._features[feature.tag]

    def reset(self):
        """
        Remove all features registered in the
        StreamFeature object.
        """
        self._features.clear()

    def tostring(self) -> str:
        """
        Return a string representation of the xml message
        """
        return self.to_bytes().decode()

    def to_bytes(self) -> bytes:
        """
        Return an encoded xml message
        """
        res = self.__copy__()
        for _, feature in self._features.items():
            res.append(feature)

        return ET.tostring(res)



from typing import Literal, Dict, Callable
from yaml import load, Loader
from xml.etree import ElementTree as ET

from pyjabber.metadata import host, config_path
from pyjabber.plugins.xep_0060.xep_0060 import PubSub, NodeAttrib
from pyjabber.stanzas.error import StanzaError as SE
from pyjabber.stanzas.IQ import IQ
from pyjabber.stream.JID import JID
from pyjabber.utils import Singleton


def iq_skeleton(element: ET.Element, disco_type: Literal['info', 'items']):
    iq_res = IQ(
        type_=IQ.TYPE.RESULT,
        id_=element.get('id'),
        from_=element.get('to'),
        to=element.get('from')
    )
    return iq_res, ET.SubElement(iq_res, 'query', attrib={'xmlns': f'http://jabber.org/protocol/disco#{disco_type}'})



class Disco(metaclass=Singleton):
    def __init__(self):
        self._handlers: Dict[str, Callable[[JID, ET.Element], bytes]] = {
            "info": self.handle_info,
            "items": self.handle_items
        }
        self._host: str = host.get()
        self._config_path: str = config_path.get()
        self._items = list(load(open(self._config_path), Loader=Loader).get('items'))

        # Search if any item has the substring pubsub, and replace the placeholder
        # with the server's host of the current session
        # A none result means the pubsub feature is disable
        self._pubsub_jid = next((s for s in list(self._items) if 'pubsub' in s), None)
        if self._pubsub_jid:
            self._pubsub_jid = self._pubsub_jid.replace('$', self._host)
            self._pubsub = PubSub()

    def feed(self, jid: JID, element: ET.Element):
        if len(element) != 1:
            return SE.invalid_xml()

        if element.find('{http://jabber.org/protocol/disco#info}query') is not None:
            return self._handlers['info'](jid, element)
        elif element.find('{http://jabber.org/protocol/disco#items}query') is not None:
            return self._handlers['items'](jid, element)
        else:
            pass

    def handle_info(self, _, element: ET.Element):
        to = element.attrib.get('to')
        if to is not None:
            to = JID(to)
        else:
            return self.server_info(element)

        if to.domain == self._host:
            return self.server_info(element)

        # Pubsub info
        if self._pubsub_jid and to == self._pubsub_jid:
            iq_res, query_res = iq_skeleton(element, 'info')

            node = self._pubsub.discover_info(element)
            if node:
                ET.SubElement(
                    query_res,
                    'identity',
                    attrib={'category': 'pubsub', 'type': node[-1]}
                )
                return ET.tostring(iq_res)

            ET.SubElement(
                query_res,
                'identity',
                attrib={'category': 'pubsub', 'type': 'service'}
            )
            ET.SubElement(query_res, 'feature', attrib={'var': 'http://jabber.org/protocol/pubsub'})
            return ET.tostring(iq_res)

    def handle_items(self, _, element: ET.Element):
        to = element.attrib.get('to')

        # Server items
        if to == self._host or to is None:
            return self.server_items(element)

        # Pubsub items
        if self._pubsub_jid and to == self._pubsub_jid:
            nodes = self._pubsub.discover_items(element)
            iq_res, query = iq_skeleton(element, 'items')
            for node in nodes:
                ET.SubElement(query, 'item', attrib={'jid': self._pubsub_jid, 'node': node[NodeAttrib.NODE.value], 'name': node[NodeAttrib.NAME.value]})
            return ET.tostring(iq_res)

    def server_info(self, element: ET.Element):
        plugins = load(open(self._config_path), Loader=Loader)['plugins']
        iq_res, query = iq_skeleton(element, 'info')
        ET.SubElement(
            query,
            'identity',
            attrib={'category': 'server', 'type': 'im', 'name': 'PyJabber Server'})

        for feature in plugins:
            ET.SubElement(query, 'feature', attrib={'var': feature})

        return ET.tostring(iq_res)

    def server_items(self, element: ET.Element):
        items = load(open(self._config_path), Loader=Loader)['items']
        iq_res, query = iq_skeleton(element, 'items')
        ET.SubElement(
            query,
            'identity',
            attrib={'category': 'server', 'type': 'im', 'name': 'PyJabber Server'})

        for i in items:
            [*keys], [*values] = zip(*i.items())
            if '$' in values[0][0]:
                jid = values[0][0].replace('$', host.get())
            else:
                jid = values[0][0]

            ET.SubElement(query, 'item', attrib={'jid': jid, 'name': keys[0]})

        return ET.tostring(iq_res)



from enum import Enum
from typing import Dict
from xml.etree import ElementTree as ET


def proceed_response() -> bytes:
    """
    Proceed Stream message.

    Indicates that the TLS upgrade process has started.
    """
    elem = ET.Element(
        "proceed", attrib={
            "xmlns": "urn:ietf:params:xml:ns:xmpp-tls"
        }
    )
    return ET.tostring(elem)


class StartTLSFeature(ET.Element):
    """
    STARTTLS Stream message.

    Indicates to the user that the server is ready for the TLS upgrade (handshake).
    """
    def __init__(
        self,
        tag: str = "starttls",
        attrib: Dict[str, str] = None,
        required: bool = True,
            **extra: str) -> None:

        default_attrib = {
            "xmlns": "urn:ietf:params:xml:ns:xmpp-tls"
        }

        super().__init__(tag, attrib or default_attrib, **extra)

        if required:
            self.append(ET.Element("required"))



from unittest.mock import MagicMock, patch
from xml.etree.ElementTree import Element
import xml.etree.ElementTree as ET
import pickle

import pytest

from pyjabber.features.presence.PresenceFeature import PresenceType
from pyjabber.stream.JID import JID
from pyjabber.utils import ClarkNotation as CN
import os
from pyjabber.stanzas.error import StanzaError as SE
from pyjabber.stream.StanzaHandler import StanzaHandler, InternalServerError


@pytest.fixture(scope='function')
def setup():
    mock_buffer = MagicMock()
    mock_buffer.get_extra_info.return_value = '127.0.0.1'

    with patch('pyjabber.stream.StanzaHandler.ConnectionManager') as MockConnectionsManager, \
         patch('pyjabber.stream.StanzaHandler.PluginManager') as MockPluginManager, \
         patch('pyjabber.stream.StanzaHandler.metadata') as mock_metadata, \
         patch('pyjabber.stream.StanzaHandler.Presence') as MockPresence, \
         patch('pyjabber.stream.StanzaHandler.logger') as mock_logger: \

        MockConnectionsManager.get_jid.return_value = 'user@localhost'
        mock_metadata.host.get.return_value = 'localhost'

        handler = StanzaHandler(mock_buffer)
        handler._functions = {
            "{jabber:client}iq": MagicMock(),
            "{jabber:client}message": MagicMock(),
            "{jabber:client}presence": MagicMock()
        }

    yield (handler, mock_buffer, MockConnectionsManager,
           MockPresence, mock_logger, mock_metadata.message_queue.get.return_value)


def test_feed_valid_element(setup):
    handler, mock_buffer, mock_connections, mock_presence, _, _ = setup
    element = Element('iq', attrib={"to": "localhost", "type": "get"})
    element.tag = "{jabber:client}iq"
    child = Element('query')
    element.append(child)

    handler.feed(element)

    handler._functions[element.tag].assert_called_once_with(element)


@patch('pyjabber.stanzas.error.StanzaError.feature_not_implemented')
def test_feed_function_key_error(mock_feature_not_implemented, setup):
    handler, mock_buffer, mock_connections, mock_presence, logger, _ = setup
    element = Element('unknown', attrib={"to": "localhost"})
    element.tag = "{jabber:client}unknown"
    child = Element('{jabber:client}query')
    element.append(child)

    try:
        handler.feed(element)
    except (KeyError, InternalServerError):
        pass  # Verify the exception was raised
    else:
        raise AssertionError("Exception not raised")

    # It shouldn't reach this part of the code
    mock_feature_not_implemented.assert_not_called()
    mock_buffer.write.assert_not_called()


def test_handleIQ(setup):
    handler, mock_buffer, mock_connections, mock_presence, _, _ = setup
    element = Element('iq', attrib={"to": "localhost", "type": "get"})
    child = Element('{jabber:client}query')  # Aseguramos que tenga el namespace adecuado
    element.append(child)

    expected_response = SE.service_unavaliable()

    with patch('pyjabber.stream.StanzaHandler.PluginManager') as MockPluginManager:
        MockPluginManager.feed.return_value = expected_response  # Simulamos la respuesta esperada
        handler._pluginManager = MockPluginManager
        handler.handle_iq(element)
        mock_buffer.write.assert_called_once_with(expected_response)


def test_handleMsg_resource(setup):
    handler, mock_buffer, mock_connections, mock_presence, _, mock_queue = setup
    element = Element('message', attrib={"to": "user@localhost/res1"})
    element.tag = "{jabber:client}message"

    mock_server_buffer = MagicMock()
    mock_connections.return_value.get_buffer_online.return_value = [(None, mock_server_buffer, [True])]

    handler.handle_msg(element)

    mock_server_buffer.write.assert_called_once_with(ET.tostring(element))
    assert mock_queue.return_value.enqueue.called is False


def test_handleMsg_no_resource_tie_priority(setup):
    handler, mock_buffer, mock_connections, mock_presence, _, mock_queue = setup
    element = Element('message', attrib={"to": "user@localhost"})
    element.tag = "{jabber:client}message"

    mock_server_buffer_1 = MagicMock()
    mock_server_buffer_2 = MagicMock()
    mock_connections.return_value.get_buffer_online.side_effect = [
        [(JID('user@localhost/res1'), mock_server_buffer_1, [True])],
        [(JID('user@localhost/res2'), mock_server_buffer_2, [True])]
    ]

    mock_presence.return_value.most_priority.return_value = [('res1', 10), ('res2', 10)]

    handler.handle_msg(element)

    mock_server_buffer_1.write.assert_called_once_with(ET.tostring(element))
    mock_server_buffer_2.write.assert_called_once_with(ET.tostring(element))
    assert mock_queue.return_value.enqueue.called is False


def test_handleMsg_enqueue_resource(setup):
    handler, mock_buffer, mock_connections, mock_presence, _, mock_queue = setup
    element = Element('message', attrib={"to": "user@localhost/res1"})
    element.tag = "{jabber:client}message"

    mock_connections.return_value.get_buffer_online.return_value = []
    mock_presence.return_value.most_priority.return_value = []

    handler.handle_msg(element)

    args = mock_queue.put_nowait.call_args.args[0]
    assert args [0] == 'MESSAGE'
    assert args [1] == 'user@localhost/res1'
    assert args [2] == ET.tostring(element)


def test_handleMsg_bare(setup):
    handler, mock_buffer, mock_connections, mock_presence, _, mock_queue = setup
    element = Element('message', attrib={"to": "user@localhost"})
    element.tag = "{jabber:client}message"

    buffer_mock_1 = MagicMock()
    buffer_mock_2 = MagicMock()

    mock_connections.return_value.get_buffer_online.side_effect = [
        [(MagicMock(), buffer_mock_1, [True])], [(MagicMock(), buffer_mock_2, [True])]
    ]
    mock_presence.return_value.most_priority.return_value = [
        ('res1', PresenceType.AVAILABLE, None, None, None),
        ('res2', PresenceType.AVAILABLE, None, None, None)
    ]

    handler.handle_msg(element)

    mock_queue.put_nowait.assert_not_called()
    assert str(mock_presence.return_value.most_priority.call_args[0][0]) == 'user@localhost'
    con_calls = mock_connections.return_value.get_buffer_online.call_args_list
    assert str(con_calls[0][0][0]) == 'user@localhost/res1'
    assert str(con_calls[1][0][0]) == 'user@localhost/res2'
    buffer_mock_1.write.assert_called_with(ET.tostring(element))
    buffer_mock_2.write.assert_called_with(ET.tostring(element))


@pytest.mark.skip
def test_handleMsg_remote_user(setup):
    handler, mock_buffer, mock_connections, mock_presence, _, _ = setup
    element = Element('message', attrib={"to": "user@remote.com"})
    element.tag = "{jabber:client}message"
    mock_connections.get_server_buffer.return_value = None
    handler.handle_msg(element)
    mock_queue_message.enqueue.assert_called_once_with('remote.com', ET.tostring(element))


def test_handlePre(setup):
    handler, mock_buffer, mock_connections, mock_presence, _, _ = setup
    element = Element('presence', attrib={"to": "localhost"})
    mock_presence.return_value.feed.return_value = 'response'
    handler.handle_pre(element)
    mock_buffer.write.assert_called_once_with('response')



from xml.etree import ElementTree as ET

from pyjabber.stanzas.IQ import IQ
from pyjabber.stream.JID import JID
from pyjabber.utils import Singleton
from pyjabber.metadata import host


class Ping(metaclass=Singleton):
    def feed(self, jid: JID, element: ET.Element):
        if element.attrib.get('to') == host.get():
            return ET.tostring(
                IQ(
                    type_=IQ.TYPE.RESULT,
                    id_=element.attrib.get('id'),
                    from_=host.get(),
                    to=str(jid)
                )
            )



import asyncio
import warnings


class StreamAlivenessMonitor:
    """
    This class is a helper to monitor the aliveness of a stream. It will call a callback if the stream is not alive after a timeout.
    """

    def __init__(self, timeout=60, callback=None):
        self._timeout = timeout
        self._timeout_callback = callback
        self._timeout_task = None
        self._reset_event = asyncio.Event()

    def __del__(self):
        self.cancel()

    async def _timeout_task_coro(self):
        try:
            await asyncio.wait_for(self._reset_event.wait(), timeout=self._timeout)
        except asyncio.TimeoutError:
            if self._timeout_callback is not None:
                self._timeout_callback()


    def reset(self):
        """
            Reset the timer. Called always after received a message from the client/server
        """
        if self._timeout_task is not None:
            self._reset_event.set()
            self._timeout_task.cancel()
        self._reset_event.clear()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self._timeout_task = asyncio.create_task(self._timeout_task_coro())

    def cancel(self):
        if self._timeout_task is not None:
            self._timeout_task.cancel()



import os
import socket
import sys
import click

if sys.platform != 'win32':
    from uvloop import run
else:
    from winloop import run

from loguru import logger
from pyjabber.server import Server

FILE_PATH = os.path.dirname(os.path.abspath(__file__))


@click.command
@click.option('--host', type=str, default='localhost',
              show_default=True, help='Host name')
@click.option('--client_port', type=int, default=5222,
              show_default=True, help='Server-to-client port')
@click.option('--server_port', type=int, default=5269,
              show_default=True, help='Server-to-server port')
@click.option('--server_out_port', type=int, default=5269,
              show_default=True, help='Server-to-server port (Out coming connection)')
@click.option('--family',
              type=click.Choice(['ipv4',
                                 'ipv6'],
                                case_sensitive=False),
              default='ipv4',
              show_default=True,
              help='(ipv4 / ipv6)')
@click.option('--timeout', type=int, default=60,
              show_default=True, help='Timeout for connection')
@click.option('--database_path', type=str, default=os.path.join(os.getcwd(), "pyjabber.db"),
              show_default=True, help='Path for database file')
@click.option('--database_purge', is_flag=True, help='Restore database file to default state (empty)')
@click.option('--database_in_memory', is_flag=True,
              help='Database in memory. The data will be erased after server shutdown')
@click.option(
              "-v",
              "--verbose",
              count=True,
              help="Show verbose debug level: -v INFO -vv DEBUG, -vvv level TRACE, ")
@click.option('--log_path', type=str, help='Path to log dumpfile')
@click.option('--debug', '-D', is_flag=True,
              help='Enables debug mode in Asyncio')

def main(
        host,
        client_port,
        server_port,
        server_out_port,
        family,
        timeout,
        database_path,
        database_purge,
        database_in_memory,
        verbose,
        log_path,
        debug):

    logger.remove()

    if log_path:
        log_file = open(os.path.join(log_path, "pyjabber.log"), 'w')
        logger.add(
            log_file,
            enqueue=True,
            format="<green>{time}</green> - <level>{level}: {message}</level>",
            level=set_verbosity(verbose),
        )

    logger.add(
        sys.stderr,
        enqueue=True,
        format="<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> - <level>{level}: {message}</level>",
        level=set_verbosity(verbose),
    )

    server = Server(
        host=host,
        client_port=client_port,
        server_port=server_port,
        server_out_port=server_out_port,
        family=socket.AF_INET if family == "ipv4" else socket.AF_INET6,
        connection_timeout=timeout,
        database_path=database_path,
        database_purge=database_purge,
        database_in_memory=database_in_memory,
    )

    run(server.start(), debug=debug)

    return 0


def set_verbosity(verbose):
    if verbose == 1:
        return 'INFO'
    elif verbose == 2:
        return 'DEBUG'
    elif verbose == 3:
        return 'TRACE'
    else:
        return 'SUCCESS'


"""Allow cookiecutter to be executable through `python -m pyjabber`."""

if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



import enum
import xml.etree.ElementTree as ET
from uuid import uuid4


class Namespaces(enum.Enum):
    """
    Defines the available namespaces in the protocol.
    """
    XMLSTREAM = "http://etherx.jabber.org/streams"
    CLIENT = "jabber:client"
    SERVER = "jabber:server"


class Stream(ET.Element):

    class Namespaces(enum.Enum):
        XMLSTREAM = "http://etherx.jabber.org/streams"
        CLIENT = "jabber:client"
        SERVER = "jabber:server"

    def __init__(
            self,
            id=None,
            from_=None,
            to=None,
            version="1.0",
            xml_lang="en",
            xmlns=Namespaces.CLIENT.value):

        if not id:
            id = str(uuid4())

        attrib = {
            k: v for k, v in (
                ("id", id),
                ("from", from_),
                ("to", to),
                ("version", version),
                ("xml:lang", xml_lang),
                ("xmlns", xmlns)) if v is not None
        }

        attrib["xmlns:stream"] = Namespaces.XMLSTREAM.value

        super().__init__("stream:stream", attrib)

    def open_tag(self) -> bytes:
        tag = f'<{self.tag}'
        for a in self.attrib:
            tag += f" {a}='{self.attrib[a]}'"
        tag += '>'
        return tag.encode()


def responseStream(attrs):
    attrs = dict(attrs)

    id = str(uuid4())
    from_ = attrs.pop((None, "from"), None)
    to = attrs.pop((None, "to"), None)
    version = attrs.pop((None, "version"), "1.0")
    lang = attrs.pop(("http://www.w3.org/XML/1998/namespace", "lang"), None)

    stream = Stream(
        id=id,
        from_=to,
        to=from_,
        version=version,
        xml_lang=lang
    )

    return stream.open_tag()



import xml.etree.ElementTree as ET
from enum import Enum
from uuid import uuid4


class IQ(ET.Element):
    """
    An IQ stanza element based on ElementTree.
    The enum TYPE define the possible values of type
    """
    class TYPE(Enum):
        GET = "get"
        RESULT = "result"
        SET = "set"
        ERROR = "error"

    def __init__(
            self,
            type_: TYPE,
            id_: str = None,
            from_: str = None,
            to: str = None,
            **extra: str) -> None:
        """
        Create an IQ stanza

        Args:
            type (IQ.TYPE): A value of IQ.TYPE enum indicating the type of the iq stanza
            id (str): Optional string identifier of the stanza
            from_ (str): The sender user
            to (str): The receiver user
        """
        attrib = {
            k: v for k, v in (
                ("id", id_),
                ("from", from_),
                ("to", to),
                ("type", type_.value)) if v is not None
        }

        if attrib.get('id') is None:
            attrib['id'] = str(uuid4())

        super().__init__('iq', attrib, **extra)



from typing import Union
from xml.etree import ElementTree as ET


class Message(ET.Element):
    def __init__(
            self,
            mto: str,
            mfrom: str,
            id: str,
            body: Union[str, ET.Element],
            mtype: Union[str, None] = "chat",
            tag: str = "message",
            **extra: str) -> None:

        attrib = {
            k: v for k, v in (
                ("id", id),
                ("from", mfrom),
                ("to", mto),
                ("type", mtype)) if v is not None
        }

        super().__init__(tag, attrib, **extra)

        if type(body) == str:
            body_elem = ET.Element("body")
            body_elem.text = body
            self.append(body_elem)

        else:
            self.append(body)



import asyncio
import threading

import pytest
import pytest_asyncio
from slixmpp import ClientXMPP

from pyjabber.server import Server


@pytest.mark.asyncio
async def test_client_connection(dummy_client):
    jid, client = dummy_client
    server = Server()

    def init_thread(loop):
        asyncio.set_event_loop(loop)
        loop.run_forever()

    thread_loop = asyncio.new_event_loop()
    thread = threading.Thread(target=init_thread, args=(thread_loop,))
    thread.start()

    server_task = asyncio.create_task(server.start())
    await server.ready.wait()

    client_future = asyncio.run_coroutine_threadsafe(client.connect(), thread_loop)

    await client.disconnected
    await asyncio.sleep(5)
    await server.stop_server()

if __name__ == "__main__":
    pytest.main()




import base64
import hashlib
from contextlib import closing
from enum import Enum
from typing import Dict, List, Tuple, Union
from uuid import uuid4
from xml.etree import ElementTree as ET

from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.metadata import host
from pyjabber.db.database import connection
from pyjabber.stanzas.error import StanzaError as SE
from pyjabber.stanzas.IQ import IQ
from pyjabber.stream.JID import JID
from pyjabber.utils import ClarkNotation as CN, Singleton


class MECHANISM(Enum):
    PLAIN = "PLAIN"
    SCRAM_SHA_1 = "SCRAM-SHA-1"
    EXTERNAL = "EXTERNAL"


class Signal(Enum):
    RESET = 0
    DONE = 1


def iq_register_result(iq_id: str) -> bytes:
    iq = ET.Element(
        "iq",
        attrib={
            "type": "result",
            "id": iq_id or str(uuid4()),
            "from": host.get()})
    return ET.tostring(iq)


class SASL(metaclass=Singleton):
    """
        SASL Class.

        An instance of this class will manage the authentication from new connections,
        during the stream negotiation process.

        :param db_connection_factory: The DB connection object used to look up the credentials

    """

    def __init__(self, db_connection_factory=connection):
        self._handlers = {
            "iq": self.handleIQ,
            "auth": self.handleAuth
        }
        self._ns = "jabber:iq:register"
        self._connection_manager = ConnectionManager()
        self._db_connection_factory = db_connection_factory
        self._peername = None

    def feed(self,
             element: ET,
             extra: Dict = None) -> Union[Tuple[Signal, bytes], bytes]:
        if extra and "peername" in extra.keys():
            self._peername = extra["peername"]

        _, tag = CN.deglose(element.tag)
        return self._handlers[tag](element)

    def handleIQ(
        self, element: ET.Element) -> Union[Tuple[Signal, bytes], bytes]:
        query = element.find("{jabber:iq:register}query")

        if query is None:
            return SE.bad_request()

        if element.attrib["type"] == IQ.TYPE.SET.value:
            new_jid = query.find("{jabber:iq:register}username")

            if new_jid is None:
                return SE.bad_request()

            new_jid = new_jid.text

            with closing(self._db_connection_factory()) as con:
                res = con.execute(
                    "SELECT * FROM credentials WHERE jid = ?", (new_jid,))
                credentials = res.fetchone()

                if credentials:
                    return SE.conflict_error(
                        element.attrib["id"])
                else:
                    pwd = query.find(
                        CN.clarkFromTuple(
                            (self._ns, "password"))).text
                    hash_pwd = hashlib.sha256(pwd.encode()).hexdigest()

                    con.execute(
                        "INSERT INTO credentials(jid, hash_pwd) VALUES (?, ?)", (new_jid, hash_pwd))
                    con.commit()
                    return iq_register_result(
                        element.attrib["id"])

        elif element.attrib["type"] == "get":
            iq = IQ(
                type_=IQ.TYPE.RESULT,
                id_=element.attrib["id"] if "id" in element.attrib.keys() else None,
            )
            query = ET.Element('{jabber:iq:register}query')
            ET.SubElement(query, "username")
            ET.SubElement(query, "password")
            iq.append(query)

            return ET.tostring(iq)

    def handleAuth(
        self, element: ET.Element) -> Union[Tuple[Signal, bytes], bytes]:
        data = base64.b64decode(element.text).split("\x00".encode())
        jid = data[1].decode()
        key_hash = hashlib.sha256(data[2]).hexdigest()

        with closing(self._db_connection_factory()) as con:
            res = con.execute(
                "SELECT hash_pwd FROM credentials WHERE jid = ?", (jid,))
            hash_pwd = res.fetchone()

        if hash_pwd and hash_pwd[0] == key_hash:
            self._connection_manager.set_jid(self._peername, JID(user=jid, domain=host.get()))
            return Signal.RESET, b"<success xmlns='urn:ietf:params:xml:ns:xmpp-sasl'/>"

        return SE.not_authorized()


def SASLFeature(mechanism_list: List[MECHANISM] = None) -> ET.Element:
    """
    SASL    Feature Stream message.

    Indicates to the client the methods available to authenticate.

    :param mechanism_list: List of Mechanism selected for the stream session. Default is PLAIN
    """
    mechanism_list = mechanism_list or [MECHANISM.PLAIN]

    element = ET.Element(
        "mechanisms",
        attrib={
            "xmlns": "urn:ietf:params:xml:ns:xmpp-sasl"
        }
    )

    for m in mechanism_list:
        mechanism = ET.SubElement(element, "mechanism")
        mechanism.text = m.value

    return element



import sqlite3

from pyjabber.metadata import database_path, database_in_memory


def connection() -> sqlite3.Connection:
    """
    Returns an already crafted connection with the database.
    It takes the parameters from the server class instance (i.e., DB path | DB in memory)
    """
    if database_in_memory.get():
        return sqlite3.connect("file::memory:?cache=shared", uri=True)
    return sqlite3.connect(database_path.get())



import xml.etree.ElementTree as ET
from contextlib import closing
from typing import List
from uuid import uuid4

from pyjabber.db.database import connection
from pyjabber.plugins.roster.Roster import Roster
from pyjabber.stream.JID import JID


def retrieve_roster(jid: str) -> List[str]:  # pragma: no cover
    with closing(connection()) as con:
        res = con.execute("SELECT * FROM roster WHERE jid = ?", (jid,))
        roster = res.fetchall()
    return roster


def update(id: int, item: ET) -> str:  # pragma: no cover
    with closing(connection()) as con:
        con.execute("UPDATE roster SET rosterItem = ? WHERE id = ?",
                    (ET.tostring(item).decode(), id))
        con.commit()
        res = con.execute("SELECT rosterItem from roster WHERE id = ?", (id,))
        res = res.fetchone()

    if res:
        return res[0]
    else:
        return ""


def store_pending_sub(from_: str, to_: str, item: ET.Element) -> None:
    with closing(connection()) as con:
        con.execute("INSERT INTO pendingsub values (?, ?, ?)", (from_, to_, ET.tostring(item).decode()))
        con.commit()


def check_pending_sub() -> List[str]:
    with closing(connection()) as con:
        res = con.execute("SELECT * FROM pendingsub")
        pending = res.fetchall()

        con.execute("DELETE FROM pendingsub")
        con.commit()
    return pending


def check_pending_sub_to(jid: str, to: str) -> ET.Element:
    with closing(connection()) as con:
        res = con.execute("SELECT * FROM pendingsub WHERE jid_from = ? AND jid_to = ?", (jid, to))
        res = res.fetchone()
    if res:
        return res


def create_roster_entry(jid: JID, to: JID, roster_manager: Roster):
    iq = ET.Element(
        "iq", attrib={"from": str(jid), "id": str(uuid4()), "type": "set"}
    )
    query = ET.Element("{jabber:iq:roster}query")
    item = ET.Element(
        "{jabber:iq:roster}item",
        attrib={
            "jid": str(to),
            "subscription": "none"})
    query.append(item)
    iq.append(query)

    return roster_manager.feed(jid, iq)




import pytest
from unittest.mock import Mock, MagicMock, patch
from xml.etree import ElementTree as ET
from base64 import b64encode
from hashlib import sha256

from pyjabber.stream.JID import JID
from pyjabber.utils import ClarkNotation as CN
from pyjabber.stream.StreamHandler import Stage, Signal, StreamHandler


@pytest.fixture
def setup():
    with patch('pyjabber.stream.StreamHandler.host') as mock_host:
        with patch('pyjabber.features.SASLFeature.host') as mock_host_sasl:
            buffer = Mock()
            starttls = Mock()
            mock_host.get.return_value = 'localhost'
            mock_host_sasl.get.return_value = 'localhost'
            yield StreamHandler(buffer, starttls)


def test_stage_enum():
    assert Stage.CONNECTED.value == 0
    assert Stage.OPENED.value == 1
    assert Stage.SSL.value == 2
    assert Stage.SASL.value == 3
    assert Stage.AUTH.value == 4
    assert Stage.BIND.value == 5
    assert Stage.READY.value == 6

def test_signal_enum():
    assert Signal.RESET.value == 0
    assert Signal.DONE.value == 1


def test_stream_handler_initialization(setup):
    handler = setup

    assert handler._buffer is not None
    assert handler._starttls is not None
    assert handler._stage == Stage.CONNECTED
    assert handler._elem is None
    assert handler._jid is None

def test_stream_handler_buffer_property(setup):
    handler = setup

    new_buffer = Mock()
    handler.buffer = new_buffer

    assert handler.buffer == new_buffer

def test_handle_open_stream_connected(setup):
    handler = setup

    handler.handle_open_stream()

    assert handler._stage == Stage.OPENED
    handler._buffer.write.assert_called_once()

def test_handle_open_stream_opened(setup):
    handler = setup
    handler._stage = Stage.OPENED

    elem = ET.Element("starttls")
    result = handler.handle_open_stream(elem)

    assert handler._stage == Stage.SSL
    assert result == Signal.RESET
    handler._buffer.write.assert_called_once()
    handler._starttls.assert_called_once()

def test_handle_open_stream_ssl(setup):
    handler = setup
    handler._stage = Stage.SSL

    handler.handle_open_stream()

    assert handler._stage == Stage.SASL
    handler._buffer.write.assert_called_once()

from pyjabber.features.SASLFeature import SASL, connection, SASLFeature


def test_handle_open_stream_sasl_continue(setup):
    handler = setup
    handler._stage = Stage.SASL

    password = b'password'
    keyhash = sha256(password).hexdigest()

    elem = ET.Element("{urn:ietf:params:xml:ns:xmpp-sasl}auth")
    auth_text = b64encode(b'\x00username\x00password').decode('ascii')
    elem.text = auth_text

    with patch.object(SASL, "__init__", lambda self: None), \
         patch.object(SASL, "feed", return_value=(Signal.RESET, b"<success xmlns='urn:ietf:params:xml:ns:xmpp-sasl'/>")):

         handler.handle_open_stream(elem)

    assert handler._stage == Stage.AUTH
    expected_response = b"<success xmlns='urn:ietf:params:xml:ns:xmpp-sasl'/>"
    handler._buffer.write.assert_called_once_with(expected_response)


def test_handle_open_stream_auth(setup):
    handler = setup
    handler._stage = Stage.AUTH

    handler.handle_open_stream()

    assert handler._stage == Stage.BIND
    handler._buffer.write.assert_called_once()

def test_handle_open_stream_bind(monkeypatch, setup):
    handler = setup
    handler._stage = Stage.BIND

    iq_elem = ET.Element("iq", attrib={"type": "set", "id": "123"})
    bind_elem = ET.SubElement(iq_elem, CN.clarkFromTuple(("urn:ietf:params:xml:ns:xmpp-bind", "bind")))
    res_elem = ET.SubElement(bind_elem, CN.clarkFromTuple(("urn:ietf:params:xml:ns:xmpp-bind", "resource")))
    res_elem.text = "resource_id"

    connections = Mock()
    handler._connection_manager = connections
    jid = JID("user@localhost")
    connections.get_jid.return_value = jid

    clark_notation_bind = CN.clarkFromTuple(("urn:ietf:params:xml:ns:xmpp-bind", "bind"))
    bind_elem_found = iq_elem.find(clark_notation_bind)
    assert bind_elem_found is not None, f"bind_elem_found should not be None, got {bind_elem_found}"

    result = handler.handle_open_stream(iq_elem)

    assert result == Signal.DONE
    handler._buffer.write.assert_called_once()

    connections.get_jid.assert_called_once_with(handler._buffer.get_extra_info('peername'))
    connections.set_jid.assert_called_once_with(
        handler._buffer.get_extra_info('peername'),
        jid,
        handler._buffer
    )

# Ejecutar los tests con pytest
if __name__ == "__main__":
    pytest.main()



from enum import Enum
from typing import List


class FieldTypes(Enum):
    BOOLEAN = 'boolean'
    FIXED = 'fixed'
    HIDDEN = 'hidden'
    JID_MULTI = 'jid-multi'
    JID_SINGLE = 'jid-single'
    LIST_MULTI = 'list-multi'
    LIST_SINGLE = 'list-single'
    TEXT_MULTI = 'text-multi'
    TEXT_PRIVATE = 'text-private'
    TEXT_SINGLE = 'text-single'

    @classmethod
    def from_value(cls, value):
        for item in cls:
            if item.value == value:
                return item
        raise ValueError()


class Option:
    def __init__(self, label: str, value: str):
        self._label = label,
        self._value = value


class FieldResponse:
    def __init__(
            self,
            field_type: FieldTypes,
            var: str,
            values: List[str] = None):

        self._type = field_type
        self._var = var
        self._values = values

    @property
    def type(self) -> FieldTypes:
        return self._type

    @property
    def var(self) -> str:
        return self._var

    @property
    def values(self) -> List[str]:
        return self._values


class FieldRequest:
    def __init__(
            self,
            field_type: FieldTypes,
            var: str,
            label: str = None,
            values: List[str] = None,
            options: List[Option] = None,
            desc: str = None,
            required: bool = False):

        if field_type not in [FieldTypes.LIST_SINGLE.value, FieldTypes.LIST_MULTI.value] and options is not None:
            raise ValueError("Options is only available with FieldType of [LIST-SINGLE, LIST-MULTIPLE]")

        if options is not None and values is not None:
            raise ValueError("Options List cannot be passed along with values List")

        self._type = field_type
        self._var = var
        self._label = label
        self._values = values
        self._options = options
        self._desc = desc
        self._required = required

    @property
    def type(self) -> FieldTypes:
        return self._type

    @property
    def label(self) -> str:
        return self._label

    @property
    def var(self) -> str:
        return self._var

    @property
    def options(self) -> List[str]:
        return self._options

    @property
    def values(self) -> List[str]:
        return self._values

    @property
    def desc(self) -> str:
        return self._desc

    @property
    def required(self) -> bool:
        return self._required



import asyncio
import os
import queue
import signal
import socket
import sqlite3
import ssl

from contextlib import closing

from loguru import logger

from pyjabber.db.database import connection
from pyjabber.network.XMLProtocol import XMLProtocol, TransportProxy
from pyjabber.network.server.outcoming.XMLServerOutcomingProtocol import XMLServerOutcomingProtocol
from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.webpage.adminPage import AdminPage
from pyjabber.network import CertGenerator
from pyjabber.metadata import database_in_memory as metadata_database_in_memory
from pyjabber import metadata
from pyjabber.workers import tls_worker, queue_worker

SERVER_FILE_PATH = os.path.dirname(os.path.abspath(__file__))


def setup_query_local_ip():
    """Return the local IP of the host machine"""
    mock_connection = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        mock_connection.connect(("8.8.8.8", 80))
        local_ip_address = mock_connection.getsockname()[0]
        return local_ip_address
    except OSError as e:
        logger.error(f"Unable to retrieve LAN IP: {e}")
    finally:
        mock_connection.close()


class Server:
    """Server class

    :param host: Host for the clients connections (localhost by default)
    :param client_port: Port for client connections (5222 by default)
    :param server_port: Port for server-to-server connections (5269 by default)
    :param family: Type of AddressFamily (IPv4 or IPv6)
    :param connection_timeout: Max time without any response from a client. After that, the server will terminate the connection
    :param database_path: Path to the sqlite3 database file. If it does not exist, a new file/database will be created (site-packages/pyjabber/db by default)
    :param database_purge: Flag to indicate the reset process of the database. CAUTION! ALL THE INFORMATION WILL BE LOST
    :param cert_path: Path to custom domain certs. By default, the server generates its own certificates for hostname
    """

    def __init__(
        self,
        host: str = "localhost",
        client_port: int = 5222,
        server_port: int = 5269,
        server_out_port: int = 5269,
        family: socket.AddressFamily = socket.AF_INET,
        connection_timeout: int = 60,
        database_path: str = os.path.join(os.getcwd(), "pyjabber.db"),
        database_purge: bool = False,
        database_in_memory: bool = False,
        cert_path: str = None
    ):
        # Server
        self._host = host
        self._client_port = client_port
        self._server_port = server_port
        self._server_out_port = server_out_port
        self._family = family
        self._client_listener = None
        self._server_listener = None
        self._adminServer = AdminPage()
        self._public_ip = None
        self._connection_timeout = connection_timeout

        # Database
        self._database_path = database_path
        self._sql_init_script = os.path.join(SERVER_FILE_PATH, "db", "schema.sql")
        self._sql_delete_script = os.path.join(SERVER_FILE_PATH, "db", "delete.sql")
        self._database_purge = database_purge
        self._database_in_memory = database_in_memory
        self._db_in_memory_con = None

        # Certs
        self._cert_path = cert_path or os.path.join(SERVER_FILE_PATH, "network", "certs")

        # Singletons
        self._connection_manager = ConnectionManager()

        # Contextvar
        metadata.host.set(host)
        metadata.database_path.set(self._database_path)
        metadata.config_path.set(os.path.join(os.path.dirname(os.path.abspath(__file__)), "config/config.yaml"))
        metadata.cert_path.set(self._cert_path)
        metadata.root_path.set(SERVER_FILE_PATH)

        # Flags
        self._ready = asyncio.Event()

    @property
    def ready(self):
        return self._ready

    def setup_database(self):
        if self._database_in_memory:
            logger.info("Using database on memory. ANY CHANGE WILL BE LOST AFTER SERVER SHUTDOWN!")
            self._db_in_memory_con = sqlite3.connect("file::memory:?cache=shared", uri=True)
            with open(self._sql_init_script, "r") as script:
                self._db_in_memory_con.cursor().executescript(script.read())
                self._db_in_memory_con.commit()
            metadata_database_in_memory.set(self._database_in_memory)

        elif os.path.isfile(self._database_path) is False:
            logger.info("No database found. Initializing one...")
            if self._database_purge:
                logger.info("Ignoring purge database flag. No DB to purge")
            with closing(connection()) as con:
                with open(self._sql_init_script, "r") as script:
                    con.cursor().executescript(script.read())
                con.commit()
        else:
            if self._database_purge:
                logger.info("Resetting the database to default state...")
                with closing(connection()) as con:
                    with open(self._sql_delete_script, "r") as script:
                        con.cursor().executescript(script.read())
                    con.commit()

    def setup_certs(self):
        try:
            if CertGenerator.check_hostname_cert_exists(self._host, self._cert_path) is False:
                CertGenerator.generate_hostname_cert(self._host, self._cert_path)
        except FileNotFoundError as e:
            logger.error(f"{e.__class__.__name__}: Pass an existing directory in your system to load the certs. "
                         f"Closing server")
            raise SystemExit

    async def setup_tls_worker(self):
        """
        A TLS Worker for process STARTTLS petitions.
        A global Asyncio queue must be declared and used across the server. The main producer of the queue will be the
        XMLProtocol class, where the transport/buffer/protocol is managed.
        The worker is global for all the server components, and it can be duplicated across multiple workers in
        different threads to handle a high number of new connections established within a very short period of time.

        """
        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_context.load_cert_chain(
            certfile=os.path.join(self._cert_path, f"{self._host}_cert.pem"),
            keyfile=os.path.join(self._cert_path, f"{self._host}_key.pem"),
        )
        loop = asyncio.get_running_loop()
        try:
            while True:
                transport, protocol, parser = await tls_queue.get()
                peer = transport.get_extra_info("peername")
                try:
                    new_transport = await loop.start_tls(
                        transport=transport.originalTransport,
                        protocol=protocol,
                        sslcontext=ssl_context,
                        server_side=True)

                    new_transport = TransportProxy(new_transport, peer)
                    transport = new_transport
                    parser.buffer = new_transport
                    logger.debug(f"Done TLS for <{peer}>")

                except ConnectionResetError:
                    logger.error(f"ERROR DURING TLS UPGRADE WITH <{peer}>")
                    if not transport.is_closing():
                        self._connection_manager.close(peer)
        except asyncio.CancelledError:
            pass

    async def run_server(self):
        """
        Launches the configured server, and returns a coroutine.
        """
        logger.info("Starting server...")

        self.setup_database()
        self.setup_certs()
        lan_ip = setup_query_local_ip()

        loop = asyncio.get_running_loop()

        try:
            self._client_listener = await loop.create_server(
                lambda: XMLProtocol(
                    namespace="jabber:client",
                    host=self._host,
                    connection_timeout=self._connection_timeout,
                    cert_path=self._cert_path,
                ),
                host=[self._host, lan_ip] if lan_ip else [self._host],
                port=self._client_port,
                family=self._family
            )
        except OSError as e:
            logger.error(e)
            self.raise_exit()

        logger.info(f"Client domain => {self._host}")
        logger.info(f"Server is listening clients on {[s.getsockname() for s in self._client_listener.sockets if s]}")

        # try:
        #     self._server_listener = await loop.create_server(
        #         lambda: XMLServerIncomingProtocol(
        #             namespace="jabber:server",
        #             host=self._host,
        #             connection_timeout=self._connection_timeout,
        #             cert_path=self._cert_path,
        #         ),
        #         host=["0.0.0.0"],
        #         port=self._server_port,
        #         family=self._family
        #     )
        # except OSError as e:
        #     logger.error(e)
        #     raise SystemExit

        # logger.info(f"Server is listening servers on {[s.getsockname() for s in self._server_listener.sockets if s]}")
        logger.success("Server started...")
        self._ready.set()

    async def stop_server(self):
        """
        Safely stops the running server
        """
        logger.info("Stopping server...")

        if self._db_in_memory_con:
            self._db_in_memory_con.close()

        if self._client_listener and self._client_listener.is_serving():
            self._client_listener.close()
            await self._client_listener.wait_closed()

        if self._server_listener and self._server_listener.is_serving():
            self._server_listener.close()
            await self._server_listener.wait_closed()

        logger.success("Server stopped...")

    # def task_s2s(self, host):
    #     # host = host.split("@")[-1]
    #     # loop = asyncio.get_running_loop()
    #     #
    #     # asyncio.ensure_future(self.server_connection(host, loop), loop=loop)

    # async def server_connection(self, remote_host, loop):
    #     """Task to connect to another XMPP server
    #     :param remote_host: Host of the remote server to connect
    #     :param loop: Asyncio running loop. Necessary to perform task
    #     """
    #     await loop.create_connection(
    #         lambda: XMLServerOutcomingProtocol(
    #             namespace="jabber:server",
    #             host=remote_host,
    #             public_host=self._public_ip,
    #             connection_timeout=self._connection_timeout,
    #         ),
    #         host=remote_host,
    #         port=self._server_out_port
    #     )

    def raise_exit(self, *args):
        """Exception used to signal the safe close of the server"""
        raise SystemExit(1)

    async def start(self):
        """Start the already created and configuration server"""
        metadata.tls_queue.set(asyncio.Queue())
        metadata.connection_queue.set(asyncio.Queue())
        metadata.message_queue.set(asyncio.Queue())

        signal.signal(signal.SIGINT, self.raise_exit)
        signal.signal(signal.SIGABRT, self.raise_exit)
        signal.signal(signal.SIGTERM, self.raise_exit)

        try:
            server_task = asyncio.create_task(self.run_server())
            tasks = [
                asyncio.create_task(self._adminServer.start()),
                asyncio.create_task(tls_worker()),
                asyncio.create_task(queue_worker())
            ]
            await asyncio.gather(*tasks)

        except (SystemExit, KeyboardInterrupt, asyncio.CancelledError) as e:  # pragma: no cover
            logger.trace(f"Signal {e.__class__.__name__} intercepted. Stopping server")

        finally:
            for task in tasks:
                task.cancel()
            try:
                await asyncio.gather(*tasks, return_exceptions=False)
                await self.stop_server()
            except (asyncio.CancelledError, SystemExit, Exception) as e:
                pass



import asyncio
import os
import queue
import ssl
from typing import Dict, List, Tuple
from xml.etree.ElementTree import Element

from loguru import logger
from xml.etree import ElementTree as ET

from pyjabber import metadata
from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.network.XMLProtocol import TransportProxy
from pyjabber.stream.JID import JID


async def tls_worker():
    """
    A TLS Worker for process STARTTLS petitions.
    A global Asyncio queue must be declared and used across the server. The main producer of the queue will be the
    XMLProtocol class, where the transport/buffer/protocol is managed.
    The worker is global for all the server components, and it can be duplicated across multiple workers in
    different threads to handle a high number of new connections established within a very short period of time.

    :param tls_queue: The global queue used ONLY for (TRANSPORT, PROTOCOL, PARSER) tuples
    :param cert_path:
    :param host:
    """
    connection_manager = ConnectionManager()
    ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ssl_context.maximum_version = ssl.TLSVersion.TLSv1_2
    ssl_context.load_cert_chain(
        certfile=os.path.join(metadata.cert_path.get(), f"{metadata.host.get()}_cert.pem"),
        keyfile=os.path.join(metadata.cert_path.get(), f"{metadata.host.get()}_key.pem"),
    )
    # ssl_context.load_cert_chain()
    loop = asyncio.get_running_loop()
    tls_queue = metadata.tls_queue.get()
    try:
        while True:
            transport, protocol, parser = await tls_queue.get()
            peer = transport.get_extra_info("peername")
            try:
                new_transport = await loop.start_tls(
                    transport=transport.originalTransport,
                    protocol=protocol,
                    sslcontext=ssl_context,
                    server_side=True)

                new_transport = TransportProxy(new_transport, peer)
                transport = new_transport
                parser.buffer = new_transport
                logger.debug(f"Done TLS for <{peer}>")

            except ConnectionResetError:
                logger.error(f"ERROR DURING TLS UPGRADE WITH <{peer}>")
                if not transport.is_closing():
                    connection_manager.close(peer)
    except asyncio.CancelledError:
        pass


async def queue_worker():
    """
       Creates a worker in the asyncio loop that accepts two different queues: one
       for new connections and another for new messages. The worker continuously
       reads from both queues.

       - For the new connections queue, the worker checks if there are any pending
         messages in the buffer for the new connection's JID. If so, it processes
         them.
       - For the new messages queue, the worker directly enqueues incoming
         messages into the pending messages buffer.

       Returns:
           None: This function does not return a value; it runs indefinitely in an
           asynchronous loop.
   """
    pending_stanzas: Dict[str, List[bytes]] = {}
    connection_manager = ConnectionManager()
    connection_queue = metadata.connection_queue.get()
    message_queue = metadata.message_queue.get()

    try:
        while True:
            con_task = asyncio.create_task(connection_queue.get())
            msg_task = asyncio.create_task(message_queue.get())

            done, pending = await asyncio.wait(
                [con_task, msg_task],
                return_when=asyncio.FIRST_COMPLETED
            )

            result = done.pop().result()

            for task in pending:
                task.cancel()

            if len(done) > 1:
                for item in done:
                    if item.result()[0] == 'CONNECTION':
                        connection_queue.put_nowait(item)
                    else:
                        message_queue.put_nowait(item)

            if result[0] == 'CONNECTION':
                _, jid = result

                if str(jid) in pending_stanzas:
                    jid = str(jid)
                elif jid.bare() in pending_stanzas:
                    jid = jid.bare()
                else:
                    jid = None

                while jid and pending_stanzas[jid]:
                    stanza_bytes = pending_stanzas[jid].pop()
                    buffer = connection_manager.get_buffer_online(JID(jid))
                    for b in buffer:
                        b[1].write(stanza_bytes)

            else:
                _, jid, stanza_bytes = result
                if str(jid) not in pending_stanzas:
                    pending_stanzas[str(jid)] = []
                pending_stanzas[str(jid)].append(stanza_bytes)

    except asyncio.CancelledError:
        pass



from enum import Enum
from loguru import logger
from xml.etree import ElementTree as ET

from pyjabber.network.XMLParser import XMLParser
from pyjabber.stream import Stream
from pyjabber.stream.StreamHandler import Signal
from pyjabber.stream.server.incoming.StanzaServerIncomingHandler import StanzaServerIncomingHandler
from pyjabber.stream.server.incoming.StreamServerIncomingHandler import StreamServerIncomingHandler
from pyjabber.utils import ClarkNotation as CN


class StreamState(Enum):
    """
    Stream connection states.
    """
    CONNECTED = 0
    READY = 1


class XMLServerIncomingParser(XMLParser):
    """
    Manages the stream data and process the XML objects.
    Inheriting from sax.ContentHandler
    """
    def __init__(self, host, buffer, starttls):
        super().__init__(host, buffer, starttls)
        self._streamHandler = StreamServerIncomingHandler(host, buffer, starttls)

    def startElementNS(self, name, qname, attrs):
        logger.debug(f"Start element NS: {name}")

        if self._stack:  # "<stream:stream>" tag already present in the data stack
            elem = ET.Element(
                CN.clarkFromTuple(name),
                attrib={CN.clarkFromTuple(key): item for key, item in dict(attrs).items()}
            )
            self._stack.append(elem)

        elif name[1] == "stream" and name[0] == "http://etherx.jabber.org/streams":
            self._buffer.write(Stream.responseStream(attrs))

            elem = ET.Element(
                CN.clarkFromTuple(name),
                attrib={CN.clarkFromTuple(key): item for key, item in dict(attrs).items()}
            )

            self._stack.append(elem)
            self._streamHandler.handle_open_stream()

        else:
            raise Exception()

    def endElementNS(self, name, qname):
        logger.debug(f"End element NS: {qname} : {name}")

        if name == "</stream:stream>":
            self._buffer.write(b'</stream>')
            self._stack.clear()
            return

        if not self._stack:
            raise Exception()

        elem = self._stack.pop()

        if elem.tag != CN.clarkFromTuple(name):
            # INVALID STANZA/MESSAGE
            raise Exception()

        if self._stack[-1].tag != '{http://etherx.jabber.org/streams}stream':
            self._stack[-1].append(elem)

        else:
            if self._state == StreamState.READY:  # Ready to process stanzas
                self._stanzaHandler.feed(elem)
            else:
                signal = self._streamHandler.handle_open_stream(elem)
                if signal == Signal.RESET and "stream" in self._stack[-1].tag:
                    self._stack.clear()
                elif signal == Signal.DONE:
                    self._stanzaHandler = StanzaServerIncomingHandler(self._buffer)
                    self._state = StreamState.READY



import asyncio
from unittest.mock import MagicMock, patch, AsyncMock
import pytest
from xml import sax
import ssl

from pyjabber.network.StreamAlivenessMonitor import StreamAlivenessMonitor
from pyjabber.network.XMLProtocol import XMLProtocol, FILE_AUTH
import os

@pytest.fixture
def setup():
    with patch('pyjabber.network.XMLProtocol.XMLParser') as mock_parser:
        with patch('pyjabber.network.XMLProtocol.ConnectionManager') as mock_connection:
            with patch('pyjabber.network.XMLProtocol.StreamAlivenessMonitor') as mock_monitor:
                with patch('pyjabber.network.XMLProtocol.Presence'):
                    with patch('pyjabber.network.XMLProtocol.metadata.tls_queue') as mock_queue:
                        protocol = XMLProtocol(
                            host='localhost',
                            namespace='jabber:client',
                            connection_timeout=30,
                            cert_path='cert_path',
                        )
                        protocol._timeout_monitor = mock_monitor
                        yield protocol, mock_parser, mock_connection, mock_monitor


@patch('pyjabber.network.XMLProtocol.sax.make_parser')
@patch('pyjabber.network.XMLProtocol.TransportProxy')
def test_connection_made(TransportProxy, make_parser, setup):
    mock_transport = MagicMock()
    protocol, parser, connection, _ = setup
    protocol.connection_made(mock_transport)

    mock_transport.get_extra_info.assert_called_with('peername')
    make_parser.assert_called_once()
    make_parser().setFeature.assert_any_call(sax.handler.feature_namespaces, True)
    make_parser().setFeature.assert_any_call(sax.handler.feature_external_ges, False)
    make_parser().setContentHandler.assert_called_once_with(parser())
    connection().connection.assert_called_once_with(mock_transport.get_extra_info(), TransportProxy())


def test_connection_lost(monkeypatch, setup):
    protocol, _, _, monitor = setup

    protocol.connection_lost(None)
    monitor.cancel.assert_called()
    assert protocol._transport is None
    assert protocol._xml_parser is None


def test_data_received(setup):
    protocol, parser, _, monitor = setup
    protocol._xml_parser = parser
    data = b"<?xml version=\'1.0\'?><stream:stream>"

    protocol.data_received(data)

    monitor.reset.assert_called_once()
    parser.feed.assert_called_once_with(b"<stream:stream>")


def test_eof_received(setup):
    protocol, _, _, _, = setup
    protocol._transport = MagicMock
    with patch('pyjabber.network.XMLProtocol.logger') as mock_logger:
        protocol.eof_received()
        mock_logger.debug.assert_called()


def test_connection_timeout(setup):
    protocol, parser, _, monitor = setup

    mock_transport = MagicMock()
    protocol._transport = mock_transport
    with patch('pyjabber.network.XMLProtocol.logger') as mock_logger:
        protocol.connection_timeout()
        mock_logger.info.assert_called()
        mock_transport.write.assert_called_once_with(b"<connection-timeout/>")
        mock_transport.close.assert_called_once()
        assert protocol._transport is None
        assert protocol._xml_parser is None


def test_enable_tls(setup):
    protocol, parser, _, _ = setup
    mock_transport = MagicMock()
    mock_handler = MagicMock()
    mock_queue = MagicMock()

    parser.getContentHandler.return_value = mock_handler
    protocol._transport = mock_transport
    protocol._xml_parser = parser
    protocol._tls_queue = mock_queue

    protocol.task_tls()

    mock_queue.put_nowait.assert_called_once()





from typing import Union
from xml.etree import ElementTree as ET

from pyjabber.stream.StreamHandler import Signal, Stage, StreamHandler


class StreamServerOutcomingHandler(StreamHandler):

    PROCEED = "{urn:ietf:params:xml:ns:xmpp-tls}proceed"
    SUCCESS = "{urn:ietf:params:xml:ns:xmpp-sasl}success"
    DIALBACK = "urn:xmpp:features:dialback{dialback}"
    MECHANISMS = "{urn:ietf:params:xml:ns:xmpp-sasl}mechanisms"
    FEATURES = "{http://etherx.jabber.org/streams}features"
    STARTTLS = "{urn:ietf:params:xml:ns:xmpp-tls}starttls"

    def __init__(self, host, buffer, starttls) -> None:
        super().__init__(host, buffer, starttls)

    def handle_open_stream(
            self, elem: ET.Element = None) -> Union[Signal, None]:
        if self._stage == Stage.READY:
            peer = self.buffer.get_extra_info('peername')
            self._connection_manager.set_server_transport(peer, self.buffer)
            return Signal.DONE

        elif elem.tag == self.FEATURES:
            children = [child.tag for child in elem]
            if self.STARTTLS in children:
                if self._stage == Stage.CONNECTED:
                    self._buffer.write(
                        "<starttls xmlns='urn:ietf:params:xml:ns:xmpp-tls'/>".encode())
                    self._stage = Stage.OPENED
                    return

            elif self.MECHANISMS in children:
                mechanisms = [mech for mech in elem.find(self.MECHANISMS)]
                if 'EXTERNAL' in [mech.text for mech in mechanisms]:
                    self._buffer.write(
                        "<auth xmlns='urn:ietf:params:xml:ns:xmpp-sasl' mechanism='EXTERNAL'>=</auth>".encode())
                    return

            elif self.DIALBACK in children and self._stage == Stage.READY:
                peer = self.buffer.get_extra_info('peername')
                self._connection_manager.set_server_transport(peer, self.buffer)
                return Signal.DONE

        elif elem.tag == self.PROCEED:
            if self._stage == Stage.OPENED:
                self._starttls()
                self._stage = Stage.SSL
                return Signal.RESET

        elif elem.tag == self.SUCCESS:
            self._stage = Stage.READY
            return Signal.RESET

        return



from xml.etree import ElementTree as ET


class InBandRegistration(ET.Element):
    """
    InBandRegistration Stream message.

    It allows the user to register itself on the server if
    there is no credentials stored with the given JID
    """
    def __init__(
        self,
        tag: str = "register",
        attrib=None,
            **extra: str) -> None:

        default_atrrib = {
            "xmlns": "http://jabber.org/features/iq-register"
        }

        super().__init__(tag, attrib or default_atrrib, **extra)



import os

from setuptools import setup, find_packages
import pyjabber


def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    with open(filename) as f:
        lineiter = [line.strip() for line in f]
    return [line for line in lineiter if line and not line.startswith("#")]


setup(
    name="pyjabber",
    version=os.environ.get("VERSION") or pyjabber.__version__,
    author="Aarón Raya Lopez, Manel Soler Sanz",
    author_email="aaron.raya.lopez@gmail.com",
    description="A Python XMPP server",
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    url="https://github.com/DinoThor/PyJabber",
    packages=find_packages(),
    install_requires=parse_requirements('requirements.txt'),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    entry_points={
        "console_scripts": [
            "pyjabber=pyjabber.__main__:main",
        ],
    },
)



class JID:
    def __init__(self, jid: str = None, user: str = None, domain: str = None, resource: str = None):
        if jid and (user or domain or resource):
            raise ValueError('You cannot pass user/domain/resource if a full jid is specified')

        if jid:
            try:
                self._user, domain = jid.split('@')
                try:
                    self._domain, self._resource = domain.split('/')
                except ValueError:
                    self._domain = domain
                    self._resource = None
            except ValueError:
                raise ValueError('Malformed JID')

        elif user and domain:
            self._user = user
            self._domain = domain
            self._resource = resource

        else:
            raise ValueError('Missing user and/or domain')

    @property
    def resource(self) -> str:
        return self._resource

    @resource.setter
    def resource(self, resource: str) -> None:
        self._resource = resource

    @property
    def domain(self) -> str:
        return self._domain

    @property
    def user(self) -> str:
        return self._user

    def bare(self) -> str:
        return f'{self._user}@{self._domain}'

    def __str__(self):
        if self._resource:
            return f'{self._user}@{self._domain}/{self._resource}'
        return f'{self._user}@{self._domain}'

    def __eq__(self, other):
        if isinstance(other, JID):
            return self.user == other.user and self.domain == other.domain and self.resource == other.resource
        return False



from xml.etree import ElementTree as ET


class ResourceBinding(ET.Element):
    """
    ResourceBinding Stream message.

    Allows the user to use a specific identifier for his session.
    If not provided, the server will automatically assign one in behalf of the client.
    """
    def __init__(self) -> None:
        super().__init__("bind", {"xmlns": "urn:ietf:params:xml:ns:xmpp-bind"})



import asyncio
import xml.etree.ElementTree as ET
from contextlib import closing
from enum import Enum
from typing import Any, Dict, Tuple, List, Union
from uuid import uuid4
from xml.etree.ElementTree import Element

from pyjabber.metadata import host, connection_queue
from pyjabber.db.database import connection
from pyjabber.network.ConnectionManager import ConnectionManager
from pyjabber.features.feature_utils import RosterUtils as RU
from pyjabber.plugins.roster.Roster import Roster
from pyjabber.stanzas.IQ import IQ
from pyjabber.stream.JID import JID
from pyjabber.utils import Singleton


class PresenceShow(Enum):
    EXTENDED_AWAY = "xa"
    AWAY = "away"
    CHAT = "chat"
    DND = "dnd"
    NONE = "none"


class PresenceType(Enum):
    AVAILABLE = "available"
    UNAVAILABLE = "unavailable"


class Presence(metaclass=Singleton):
    def __init__(self) -> None:
        self._handlers = {
            "subscribe": self.handle_subscribe,
            "subscribed": self.handle_subscribed,
            "unsubscribed": self.handle_unsubscribed,
            "unavailable": self.handle_global_presence,
            "INTERNAL": self.handle_lost_connection
        }
        self._connections = ConnectionManager()
        self._roster = Roster()

        self._pending = {}
        self._get_all_pending_presence()
        self._connection_queue: asyncio.Queue = connection_queue.get()

        self._online_status = {}

    def _get_all_pending_presence(self):
        with closing(connection()) as con:
            res = con.execute("SELECT jid, item FROM pendingsub", ())
            res = res.fetchall()
        for jid_from, jid_to, item in res:
            if jid_to not in self._pending:
                self._pending[jid_to] = [item]
            else:
                self._pending[jid_to].append(item)

    def _update_pending_presence(self):
        pass

    def _delete_pending_presence(self, jid: str):
        with closing(connection()) as con:
            con.execute("DELETE FROM pendingsub WHERE jid = ?", (jid,))
            con.commit()
        self._pending[jid].pop()

    def priority_by_jid(self, jid: JID):
        try:
            return self._online_status[jid.bare()]
        except KeyError:
            return []

    def most_priority(self, jid: JID) -> List[Tuple[str, PresenceType, Union[str, None], Union[str, None], Union[str, None]]]:
        priority = self.priority_by_jid(jid)

        values = [item[-1] for item in priority if item[-1] is not None]
        max_value = max(values) if values else None
        return [item for item in priority if item[-1] == max_value]

    def feed(self, jid: JID, element: Element, extra: Dict[str, Any] = None):
        if "type" in element.attrib:
            return self._handlers[element.get("type")](jid, element)

        if "to" not in element.attrib:
            return self.handle_global_presence(jid, element)

        return None

    def handle_lost_connection(self, jid: JID, element: Element):
        if jid.bare() not in self._online_status:
            return None

        index, present = self._present_in_online_list(jid)
        if present:
            if self._online_status[jid.bare()][index][1] == PresenceType.UNAVAILABLE:
                self._online_status[jid.bare()].pop(index)
                if len(self._online_status[jid.bare()]) == 0:
                    self._online_status.pop(jid.bare())
                return None

        for contact in self._roster.roster_by_jid(jid):
            item = ET.fromstring(contact.get("item"))
            if item.attrib.get("subscription") not in ["from", "both"]:
                continue

            contact_jid = item.attrib.get("jid")

            if len(contact_jid.split("@")) < 2:
                contact_jid = JID(contact_jid + f"@{host.get()}")
            else:
                contact_jid = JID(contact_jid)

            if contact_jid.bare() in self._online_status:
                for user_connected in [i for i in self._online_status[contact_jid.bare()] if
                                       i[1] == PresenceType.AVAILABLE]:
                    resource = user_connected[0]
                    dest_jid, buffer, _ = self._connections.get_buffer(
                        JID(user=contact_jid.user, domain=contact_jid.domain, resource=resource))[0]

                    presence = ET.Element(
                        "presence",
                        attrib={
                            "from": str(jid),
                            "to": dest_jid.bare(),
                            "type": PresenceType.UNAVAILABLE.value
                        }
                    )
                    buffer.write(ET.tostring(presence))

        for _, neighbour_resource, _ in self._connections.get_buffer(JID(jid.bare())):
            presence = ET.Element(
                "presence",
                attrib={
                    "from": str(jid),
                    "to": jid.bare(),
                    "type": PresenceType.UNAVAILABLE.value
                }
            )
            neighbour_resource.write(ET.tostring(presence))

    def handle_global_presence(self, jid: JID, element: ET.Element):
        if jid.bare() not in self._online_status:
            self._online_status[jid.bare()] = []

        show = next((c.text for c in element if c.tag == 'show' and c.text in PresenceShow), None)
        status = next((c.text for c in element if c.tag == 'status'), None)
        priority = next((c.text for c in element if c.tag == 'priority'), None)

        index, present = self._present_in_online_list(jid)
        if element.attrib.get("type") == PresenceType.UNAVAILABLE.value:
            if present:
                self._online_status[jid.bare()][index] = (
                    jid.resource, PresenceType.UNAVAILABLE, show, status, priority
                )
            else:
                self._online_status[jid.bare()].append(
                    (jid.resource, PresenceType.UNAVAILABLE, show, status, priority)
                )

            self._connections.online(jid, False)
        else:
            if present:
                previous_status = self._online_status[jid.bare()][index][1]
                self._online_status[jid.bare()][index] = (jid.resource, PresenceType.AVAILABLE, show, status, priority)
                if previous_status == PresenceType.UNAVAILABLE:
                    self._connection_queue.put_nowait(('CONNECTION', jid))
            else:
                self._online_status[jid.bare()].append((jid.resource, PresenceType.AVAILABLE, show, status, priority))
                self._connection_queue.put_nowait(('CONNECTION', jid))

            self._connections.online(jid)

        for contact in self._roster.roster_by_jid(jid):
            item = ET.fromstring(contact.get("item"))
            if item.attrib.get("subscription") not in ["from", "both"]:
                continue

            contact_jid = item.attrib.get("jid")

            if len(contact_jid.split("@")) < 2:
                contact_jid = JID(contact_jid + f"@{host.get()}")
            else:
                contact_jid = JID(contact_jid)

            if contact_jid.bare() in self._online_status:
                for user_connected in [i for i in self._online_status[contact_jid.bare()] if i[1] == PresenceType.AVAILABLE]:
                    resource = user_connected[0]
                    online = self._connections.get_buffer(JID(user=contact_jid.user, domain=contact_jid.domain, resource=resource))
                    if online:
                        dest_jid, buffer, _ = online[0]

                        presence = ET.Element(
                            "presence",
                            attrib={
                                "from": str(jid),
                                "to": dest_jid.bare()
                            }
                        )
                        for attrib in user_connected[2:]:
                            if attrib:
                                presence.append(attrib)

                        buffer.write(ET.tostring(presence))

        if jid.bare() in self._pending:
            for item in self._pending[jid.bare()]:
                for buffer in self._connections.get_buffer(jid):
                    buffer[1].write(item)
            self._delete_pending_presence(jid.bare())

        return None

    def handle_subscribe(self, jid: JID, element: ET.Element):
        to = JID(element.attrib.get("to"))

        if "from" not in element.attrib:
            element.attrib["from"] = str(jid)

        # Handle local presence. Receiver client connected to the server
        if to.domain == host.get():
            roster = self._roster.roster_by_jid(jid)

            item = [item for item in roster
                    if ET.fromstring(item.get("item")).attrib.get("jid") == to.user
                    or ET.fromstring(item.get("item")).attrib.get("jid") == to.bare()
                    ]

            if not item:
                self._roster.create_roster_entry(jid, to)
                roster = self._roster.roster_by_jid(jid)
                item = [item for item in roster
                        if ET.fromstring(item.get("item")).attrib["jid"] == to.user
                        or ET.fromstring(item.get("item")).attrib["jid"] == to.bare()
                        ]

            roster_item = item[0].get("item")
            et_item = ET.fromstring(roster_item)
            item_id = item[0].get("id")

            if et_item.attrib.get("ask") == "subscribe":
                return

            if et_item.attrib.get("subscription") in ["to", "both"]:
                petition = ET.Element(
                    "presence",
                    attrib={
                        "from": element.attrib.get('to'),
                        "to": element.attrib.get('from'),
                        "id": element.attrib.get('id') or str(uuid4()),
                        "type": "subscribed"
                    }
                )
                return ET.tostring(petition)

            if "ask" not in et_item.attrib:
                new_item = et_item.__copy__()
                new_item.attrib["ask"] = "subscribe"
                self._roster.update_item(new_item, item_id)

            buffer = self._connections.get_buffer(to)
            if buffer:
                petition = ET.Element(
                    "presence",
                    attrib={
                        "from": element.attrib.get('from'),
                        "to": element.attrib.get('to'),
                        "id": element.attrib.get('id') or str(uuid4()),
                        "type": "subscribe"
                    }
                )
                for b in buffer:
                    b[1].write(ET.tostring(petition))

            else:
                self._roster.store_pending_sub(to.bare(), element)

    def handle_subscribed(self, jid: JID, element: ET.Element):
        to = JID(element.attrib.get("to"))

        if "from" not in element.attrib:
            element.attrib["from"] = str(jid)

        # Handle local presence. Receiver client connected to the server
        if to.domain == host.get():
            roster_sender = self._roster.roster_by_jid(to)
            roster_receiver = self._roster.roster_by_jid(jid)

            if not roster_sender:
                return None

            if not roster_receiver:
                self._roster.create_roster_entry(JID(jid.bare()), to)
                roster_receiver = self._roster.roster_by_jid(jid)

            roster_push_sender = None
            roster_push_receiver = None

            item_sender = [  # Sender appears in receiver roster
                item for item in roster_sender
                if ET.fromstring(item.get("item")).attrib.get("jid") == jid.user
            ]

            item_receiver = [  # Receiver appears in sender roster
                item for item in roster_receiver
                if ET.fromstring(item.get("item")).attrib.get("jid") == to.user
            ]

            if item_sender:
                item_id = item_sender[0].get("id")
                item_sender = item_sender[0].get("item")
                et_item_sender = ET.fromstring(item_sender)

                if et_item_sender.attrib.get("subscription") == "none":
                    new_item_sender = et_item_sender.__copy__()
                    if "ask" in new_item_sender.attrib:
                        new_item_sender.attrib.pop("ask")

                    new_item_sender.attrib["subscription"] = "to"
                    self._roster.update_item(new_item_sender, item_id)

                    new_item_sender.attrib['jid'] = new_item_sender.attrib['jid'] + f"@{host.get()}"
                    roster_push_sender = new_item_sender

                elif et_item_sender.attrib.get("subscription") == "from":
                    new_item_sender = et_item_sender.__copy__()
                    if "ask" in new_item_sender.attrib:
                        new_item_sender.attrib.pop("ask")

                    new_item_sender.attrib["subscription"] = "both"
                    self._roster.update_item(new_item_sender, item_id)

                    new_item_sender.attrib['jid'] = new_item_sender.attrib['jid'] + f"@{host.get()}"
                    roster_push_sender = new_item_sender

                else:
                    item_sender = None

            if item_receiver:
                item_id = item_receiver[0].get("id")
                item_receiver = item_receiver[0].get("item")
                et_item_receiver = ET.fromstring(item_receiver)

                if et_item_receiver.attrib.get("subscription") == "none":
                    new_item_receiver = et_item_receiver.__copy__()
                    new_item_receiver.attrib["subscription"] = "from"
                    self._roster.update_item(new_item_receiver, item_id)

                    new_item_receiver.attrib['jid'] = new_item_receiver.attrib['jid'] + f"@{host.get()}"
                    roster_push_receiver = new_item_receiver

                elif et_item_receiver.attrib["subscription"] == "to":
                    new_item_receiver = et_item_receiver.__copy__()
                    new_item_receiver.attrib["subscription"] = "both"
                    self._roster.update_item(new_item_receiver, item_id)

                    new_item_receiver.attrib['jid'] = new_item_receiver.attrib['jid'] + f"@{host.get()}"
                    roster_push_receiver = new_item_receiver

                else:
                    item_receiver = None

            if item_receiver:
                for sender in self._connections.get_buffer(JID(to.bare())):
                    res = ET.Element(
                        "presence",
                        attrib={
                            "from": jid.bare(),
                            "to": str(to),
                            "id": element.attrib.get('id') or str(uuid4()),
                            "type": "subscribed",
                        },
                    )
                    sender[1].write(ET.tostring(res))
                return None

            if roster_push_sender is not None:
                for sender in self._connections.get_buffer(JID(to.bare())):
                    res = IQ(to=str(sender[0]), type_=IQ.TYPE.SET)
                    query = ET.SubElement(res, "query", attrib={"xmlns": "jabber:iq:roster"})
                    query.append(roster_push_sender)

                    sender[1].write(ET.tostring(res))

            if roster_push_receiver is not None:
                for receiver in self._connections.get_buffer(JID(jid.bare())):
                    res = IQ(to=str(receiver[0]), type_=IQ.TYPE.SET)
                    query = ET.SubElement(res, "query", attrib={"xmlns": "jabber:iq:roster"})
                    query.append(roster_push_receiver)

                    receiver[1].write(ET.tostring(res))

    def handle_unsubscribed(self, jid: JID, element: ET.Element):
        to = JID(element.attrib["to"])
        # bare_jid = self._jid.bare()

        if "from" not in element.attrib:
            element.attrib["from"] = str(jid)

        # Handle locally
        if to.domain == host.get():
            roster = self._roster.roster_by_jid(jid)
            # buffer =

            item = [item for item in roster
                    if ET.fromstring(item.get("item")).attrib.get("jid") in [to.bare(), to.user]]

            if not item:
                return

            item_id = item[0].get("id")
            item = item[0].get("item")
            et_item = ET.fromstring(item)
            new_item = et_item.__copy__()

            updated = False

            if et_item.attrib.get("subscription") == "to":
                new_item.attrib["subscription"] = "none"
                self._roster.update_item(item=new_item, id_=item_id)
                updated = True

            elif et_item.attrib.get("subscription") == "both":
                new_item.attrib["subscription"] = "from"
                self._roster.update_item(item=new_item, id_=item_id)
                updated = True

            elif "ask" in et_item.attrib:
                new_item.attrib.pop("ask")
                self._roster.update_item(item=new_item, id_=item_id)

            if updated:
                presence = ET.Element(
                    "presence",
                    attrib={
                        "from": element.attrib.get('from'),
                        "to": to.bare(),
                        "id": str(uuid4()),
                        "type": "unsubscribed"
                    }
                )
                for buffer in self._connections.get_buffer(to):
                    buffer[1].write(ET.tostring(presence))
                # roster_push = ET.Element(
                #     "iq",
                #     attrib={
                #         "id": id_iq,
                #         "to": b[0],
                #         "type": "set"
                #     }
                # )
                #
                # query = ET.Element("{jabber:iq:roster}query")
                # query.append(new_item)
                # roster_push.append(query)
                #
                # b[1].write(ET.tostring(presence))
                # b[1].write(ET.tostring(roster_push))

    def handle_unavailable(self, jid: JID, element: ET.Element):
        # bare_jid = self._jid.bare()

        if "from" not in element.attrib:
            element.attrib["from"] = str(jid)

        roster = RU.retrieve_roster(jid.bare())

        for item in roster:
            to = ET.fromstring(item[-1]).attrib.get('jid')

            if to.split("@")[1] == host.get():
                presence = element.__copy__()

                presence.attrib["to"] = to

                buffer = self._connections.get_buffer(JID(to))
                for b in buffer:
                    b[1].write(ET.tostring(presence))

    def _present_in_online_list(self, jid: JID):
        if jid.bare() not in self._online_status:
            return None, None

        present = [(i,p) for i, p in enumerate(self._online_status[jid.bare()]) if p[0] == jid.resource]
        if present:
            return present[0]
        return None, None
